package bingo.base;

import java.awt.BorderLayout;
import java.awt.Color;
import java.awt.Container;
import java.awt.Font;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.util.*;
import java.text.*;

import javax.swing.JButton;
import javax.swing.JFrame;
import javax.swing.JInternalFrame;
import javax.swing.JLabel;
import javax.swing.JScrollPane;
import javax.swing.JTable;
import javax.swing.tree.DefaultMutableTreeNode;

import bingo.language.LanguageConstants;
import bingo.svmlight.Feature;
import bingo.util.*;

/**
 *  Klasse zur Ausgabe eines Featurevektors zu einer URL.
 *  Es werden zu den Dokumenten jeweils die Terme und die jeweiligen relativen Hufigkeiten des Auftretens
 *  der Terme ausgegeben.
 */
public class FeatureViewer extends JFrame implements ActionListener
{
   
   private static LanguageConstants lc = LanguageConstants.getInstance();
   
   private SessionBuffer session;
   private JLabel urlLabel;
   private JButton ok = new JButton( lc.get( "ok" ) );
   private URLClassTableModel featureTable;
   private final JTable table;
   
   Font f = new Font( "Serif", Font.PLAIN, 14 );
   Font ff = new Font( "Serif", Font.ITALIC, 18 );
   DefaultMutableTreeNode parent = null;
   
   
   /**
    *  Konstruktor fr das FeatureViewer Objekt eines BingoDocuments
    */
   public FeatureViewer( JInternalFrame parent, BingoDocument bd)
   {
      if (bd == null)
      {
         session.log("BingoDocument ist null", 1);
      }
      
      this.setTitle( lc.get( "featureViewerTitle" ) + ": " );
      this.setSize( 400, 450 );
      this.setLocationRelativeTo(parent);
      addWindowListener( new CloseWindow( this ) );
      Container content = this.getContentPane();
      
      String[] columnNames =
      {"Feature", "RTF", "Weight"};
      
      urlLabel = new JLabel( bd.getUrl() );
      urlLabel.setFont( f );
      featureTable =
      new URLClassTableModel()
      {
         public boolean isCellEditable( int x, int y )
         {
            return false;
         }
      };
      
      TableSorter sorter = new TableSorter( featureTable );
      sorter.sortByColumn(0, true);
      featureTable.setColumnNames( columnNames );
      table = new JTable( sorter );
      
      sorter.addMouseListenerToHeaderInTable( table );
      
      table.setBackground( Color.gray );
      table.setForeground( Color.white );
      
      JScrollPane scrollpane = new JScrollPane( table );
      
      content.add( urlLabel, BorderLayout.NORTH );
      content.add( scrollpane, BorderLayout.CENTER );
      content.add( ok, BorderLayout.SOUTH );
      
      // ActionListener fr alle Mens registrieren
      ok.addActionListener( this );
      buildTable( bd );
      sorter.sort(featureTable);
   }
   
   public FeatureViewer( JInternalFrame parent, String sURL, String sClass, ArrayList feat )
   {
      session = SessionBuffer.getInstance();
      //rootNode = session.getRootNode();
      
      this.setTitle( lc.get( "featureViewerTitle" ) + ": " );
      this.setSize( 400, 450 );
      this.setLocationRelativeTo(parent);
      addWindowListener( new CloseWindow( this ) );
      
      // Aufbau der GUI
      Container content = this.getContentPane();
      
      String[] columnNames =
      {"TERM", "RTF", "Weight"};
      
      urlLabel = new JLabel( sURL + " - " + sClass );
      urlLabel.setFont( f );
      featureTable =
      new URLClassTableModel()
      {
         public boolean isCellEditable( int x, int y )
         {
            return false;
         }
      };
      
      TableSorter sorter = new TableSorter( featureTable );
      sorter.sortByColumn(0, true);
      featureTable.setColumnNames( columnNames );
      table = new JTable( sorter );
      
      sorter.addMouseListenerToHeaderInTable( table );
      
      table.setBackground( Color.gray );
      table.setForeground( Color.white );
      
      JScrollPane scrollpane = new JScrollPane( table );
      
      content.add( urlLabel, BorderLayout.NORTH );
      content.add( scrollpane, BorderLayout.CENTER );
      content.add( ok, BorderLayout.SOUTH );
      
      // ActionListener fr alle Mens registrieren
      ok.addActionListener( this );
      buildTable( sClass, feat );
      sorter.sort(featureTable);
   }
   
   
   private void buildTable( BingoDocument bd )
   {
      if (bd == null)
         return;
      
      Collection features = bd.getFeatureVector();
      if (features == null || features.size() <= 0)
         return;

      BingoTreeNode myNode = bd.getNode();
      HashMap mappedFeatures = new HashMap(features.size());
      if (myNode != null && !myNode.isRoot())  // root-Node hat keinen Classifier 
      {   
        bd.setMappedFeatures( myNode.documentMapping(bd.getFeatureVector()));
        Object[] featuresForClass =
        ((BingoTreeNode) myNode.getParent()).classMapping(bd.getMappedFeatures());   
        for (int i=0; i<featuresForClass.length; i++)
        {
          Feature f = (Feature)featuresForClass[i]; 
          mappedFeatures.put(f.getTerm(), f);
        }   
        bd.setMappedFeatures(null);
        featuresForClass = null;
      }
      
      DecimalFormat formatter = new DecimalFormat("#0.#####");
      
      Iterator it = features.iterator();
      while (it.hasNext())
      {
         Feature feature = (Feature)it.next();
         Object[] sRow = new Object[3];
         sRow[0] = feature.getTerm();
         sRow[1] = formatter.format(feature.getTF());
         
         if (myNode != null)
         {
            Integer termID = (Integer) (myNode.getTermID( (String)feature.getTerm() ) );
            if (termID != null)
            {
              Feature ff = (Feature) mappedFeatures.get(termID);
              if (ff != null)
                sRow[2] = formatter.format( ff.getTF());
              else
                sRow[2] = "not used";
            }   
            else
            {
              sRow[2] = "not used";
            }
         }   
         
         featureTable.addRow( sRow );
      }
      it=null;
      formatter = null;
      features = null;
      
      this.setTitle( lc.get( "featureViewerTitle" ) + ": (" + featureTable.getDataVector().size() + " " + lc.get( "fVRows" ) + ")" );
   }
   
   private void buildTable( String sClass, ArrayList features )
   {
      if (features == null || sClass==null)
      {
         return;
      }
      try
      {
         DecimalFormat formatter = new DecimalFormat("#0.#####");
         /*
         NumberFormat nf = NumberFormat.getInstance( Locale.US );
         nf.setMaximumFractionDigits( 4 );
          
         double TF;
         double IDF = 0.0;
         double TF_IDF = 0.0;
         double MI = 0.0;
          */
         Iterator it = features.iterator();
         while (it.hasNext())
         {
            Feature feature = (Feature)it.next();
            Object[] sRow = new Object[3];
                       
            sRow[0] = feature.getTerm();
            
            featureTable.addRow( sRow );
         }
         it=null;
         this.setTitle( lc.get( "featureViewerTitle" ) + ": (" + featureTable.getDataVector().size() + " " + lc.get( "fVRows" ) + ")" );
      }
      catch ( Exception any )
      {
         session.log(any);
      }
   }
   
   /**
    *  berwacht den Button.
    *
    * @param  e  Das Event.
    */
   public void actionPerformed( ActionEvent e )
   {
      String cmd = e.getActionCommand();
      
      if ( cmd == ( lc.get( "ok" ) ) )
      {
         super.dispose();
      }
   }
}
