package bingo.crawler;

import java.util.Enumeration;
import java.util.Hashtable;
import java.util.StringTokenizer;

import bingo.util.ArrayEnumeration;
import bingo.util.Str;

/**
 *  An Object with Lables.
 *
 * @author     martin
 * @created    21. Mrz 2002
 */
public class LabledObject {

  public final static String TRUE = "true".intern();

  final static int INITIAL_SIZE = 4;

  protected Hashtable names = null;


  /**
   *  Set an object-valued label.
   *
   * @param  name   name of label (case-sensitive, whitespace permitted)
   * @param  value  value set for label. If null, the label is removed.
   */
  public void setObjectLabel( String name, Object value ) {
    if ( value == null ) {
      removeLabel( name );
    } else {
      if ( names == null ) {
        names = new Hashtable( INITIAL_SIZE );
      }
      names.put( name, value );
    }
  }


  /**
   *  Set a string-valued label.
   *
   * @param  name   name of label (case-sensitive, whitespace permitted)
   * @param  value  value set for label. If null, the label is removed.
   */
  public void setLabel( String name, String value ) {
    setObjectLabel( name, value );
  }


  /**
   *  Set a label on the region. The value of the label defaults to TRUE.
   *
   * @param  name  name of label (case-sensitive, whitespace permitted)
   */
  public void setLabel( String name ) {
    setObjectLabel( name, TRUE );
  }


  /**
   *  Get an object-valued label.
   *
   * @param  name  name of label (case-sensitive, whitespace permitted)
   * @return       Object value set for label, or null if label not set
   */
  public Object getObjectLabel( String name ) {
    return names != null ? names.get( name ) : null;
  }


  /**
   *  Get a String containing the labels of the region.
   *
   * @return    string containing the label names, separated by spaces
   */
  public String getObjectLabels() {
    Enumeration enum = enumerateObjectLabels();
    StringBuffer buf = new StringBuffer();

    while ( enum.hasMoreElements() ) {
      if ( buf.length() > 0 ) {
        buf.append( ' ' );
      }
      buf.append( (String) enum.nextElement() );
    }
    return buf.toString();
  }


  /**
   *  Get a label's value.
   *
   * @param  name  name of label (case-sensitive, whitespace permitted)
   * @return       value of label, or null if label not set
   */
  public String getLabel( String name ) {
    Object obj = getObjectLabel( name );

    if ( obj == null ) {
      return null;
    } else {
      return obj.toString();
    }
  }


  /**
   *  Get a label's value. If the label is not set, return defaultValue.
   *
   * @param  name          name of label (case-sensitive, whitespace permitted)
   * @param  defaultValue  default value that should be returned if label is not
   *      set
   * @return               value of label, or defaultValue if not set
   */
  public String getLabel( String name, String defaultValue ) {
    String val = getLabel( name );

    return ( val != null ) ? val : defaultValue;
  }


  /**
   *  Get a label's value as a number. Returns the first number (integral or
   *  floating point) that can be parsed from the label's value, skipping an
   *  arbitrary amount of junk.
   *
   * @param  name          name of label (case-sensitive, whitespace permitted)
   * @param  defaultValue  default value that should be returned if label is not
   *      set
   * @return               numeric value of label, or defaultValue if not set or
   *      no number is found
   */
  public Number getNumericLabel( String name, Number defaultValue ) {
    String val = getLabel( name );

    if ( val == null ) {
      return defaultValue;
    }
    try {
      return Str.parseNumber( val );
    }
    catch ( NumberFormatException e ) {
      return defaultValue;
    }
  }


  /**
   *  Enumerate the labels of the region.
   *
   * @return    enumeration producing label names
   */
  public Enumeration enumerateObjectLabels() {
    return names != null ? names.keys() : new ArrayEnumeration( null );
  }


  /**
   *  Test if a label is set.
   *
   * @param  name  name of label (case-sensitive, whitespace permitted)
   * @return       true if label is set, otherwise false
   */
  public boolean hasLabel( String name ) {
    return names != null && names.containsKey( name );
  }


  /**
   *  Test if one or more of several labels are set.
   *
   * @param  expr  a list of label names separated by spaces
   * @return       true if region has at least one of the labels in expr
   */
  public boolean hasAnyLabels( String expr ) {
    StringTokenizer tok = new StringTokenizer( expr );

    while ( tok.hasMoreElements() ) {
      if ( hasLabel( tok.nextToken() ) ) {
        return true;
      }
    }
    return false;
  }


  /**
   *  Test if one or more of several labels are set.
   *
   * @param  labels  an array of label names
   * @return         true if region has at least one of the labels
   */
  public boolean hasAnyLabels( String[] labels ) {
    for ( int i = 0; i < labels.length; ++i ) {
      if ( hasLabel( labels[i] ) ) {
        return true;
      }
    }
    return false;
  }


  /**
   *  Test if all of several labels are set.
   *
   * @param  expr  a list of label names separated by spaces
   * @return       true if region has at least one of the labels in expr
   */
  public boolean hasAllLabels( String expr ) {
    StringTokenizer tok = new StringTokenizer( expr );

    while ( tok.hasMoreElements() ) {
      if ( !hasLabel( tok.nextToken() ) ) {
        return false;
      }
    }
    return true;
  }


  /**
   *  Test if all of several labels are set.
   *
   * @param  labels  an array of label names
   * @return         true if region has all of the labels
   */
  public boolean hasAllLabels( String[] labels ) {
    for ( int i = 0; i < labels.length; ++i ) {
      if ( !hasLabel( labels[i] ) ) {
        return false;
      }
    }
    return true;
  }


  /**
   *  Remove a label.
   *
   * @param  name  name of label (case-sensitive, whitespace permitted)
   */
  public void removeLabel( String name ) {
    if ( names != null ) {
      names.remove( name );
    }
  }
}

