package bingo.crawler;

import bingo.util.BingoDocument;

/**
 *  Link event. A LinkEvent is issued when the crawler starts or stops
 *  retrieving a link, and when it makes a decision about a link.
 *
 * @author     T
 * @created    15. Juni 2001
 */
public class LinkEvent {
  URL_Crawler crawler;
  int id;
  BingoDocument bd;
  Throwable exception;

  /**
   *  No event occured on this link yet. Never delivered in a LinkEvent, but may
   *  be returned by link.getStatus().
   */
  public final static int NONE = 0;

  /**
   *  Link was rejected by shouldVisit()
   */
  public final static int SKIPPED = 1;

  /**
   *  Link has already been visited during the crawl, so it was skipped.
   */
  public final static int ALREADY_VISITED = 2;

  /**
   *  Link was accepted by walk() but exceeds the maximum depth from the start
   *  set.
   */
  public final static int TOO_DEEP = 3;

  /**
   *  Link was accepted by walk() and is waiting to be downloaded
   */
  public final static int QUEUED = 4;

  /**
   *  Link is being retrieved
   */
  public final static int RETRIEVING = 5;

  /**
   *  An error occurred in retrieving the page. The error can be obtained from
   *  getException().
   */
  public final static int ERROR = 6;

  /**
   *  Link has been retrieved
   */
  public final static int DOWNLOADED = 7;

  /**
   *  Link has been thoroughly processed by crawler
   */
  public final static int VISITED = 8;

  /**
   *  Map from id code (RETRIEVING) to name ("retrieving")
   */
  public static String[] eventName = {
      "none",
      "skipped",
      "already visited",
      "too deep",
      "queued",
      "retrieving",
      "error",
      "downloaded",
      "visited"
      };


  /**
   *  Make a LinkEvent.
   *
   * @param  crawler  Crawler that generated this event
   * @param  id       event code, like LinkEvent.RETRIEVING
   * @param  link     Link on which this event occurred
   */
  public LinkEvent( URL_Crawler crawler, int id, BingoDocument bd ) {
    this.crawler = crawler;
    this.id = id;
    this.bd = bd;
  }


  /**
   *  Make a LinkEvent for an error.
   *
   * @param  crawler    Crawler that generated this event
   * @param  id         Event code, usually ERROR
   * @param  link       Link on which this event occurred
   * @param  exception  Throwable
   */
  public LinkEvent( URL_Crawler crawler, int id, BingoDocument bd, Throwable exception ) {
    this.crawler = crawler;
    this.id = id;
    this.bd = bd;
    this.exception = exception;
  }


  /**
   *  Get crawler that generated the event
   *
   * @return    crawler
   */
  public URL_Crawler getCrawler() {
    return crawler;
  }


  /**
   *  Get event id
   *
   * @return    id
   */
  public int getID() {
    return id;
  }


  /**
   *  Get event name (string equivalent to its ID)
   *
   * @return    id
   */
  public String getName() {
    return eventName[id];
  }


  /**
   *  Get link to which this event occurred.
   *
   * @return    link
   */
  public BingoDocument getLink() {
    return bd;
  }


  /**
   *  Get exception related to this event. Valid when ID == ERROR.
   *
   * @return    exception object
   */
  public Throwable getException() {
    return exception;
  }


  /**
   *  Convert this event to a String describing it.
   *
   * @return    Description of the Returned Value
   */
  public String toString() {
    String result;

    if ( id == ERROR ) {
      result = exception.toString();
    } else {
      result = eventName[id];
    }
    result += " " + bd.toDescription();
    return result;
  }
}

