package bingo.svmlight;

import java.io.BufferedReader;
import java.io.BufferedWriter;
import java.io.File;
import java.io.FileWriter;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Enumeration;
import java.util.Iterator;

import org.biojava.stats.svm.SVMClassifierModel;
import org.biojava.stats.svm.tools.SVM_Light;

import bingo.db.DBInterface;
import bingo.util.BingoDocument;
import bingo.util.BingoTreeNode;
import bingo.util.SessionBuffer;

/**
 *  Diese Klasse dient dazu, zu einer vorgegebenen Baumstruktur (mit den darin enthaltenen Dokumenten)
 *  SVM-Modelle zu generieren. Diese werden zur Klassifikation von neuen Dokumenten
 *  bentigt.
 */
public class SVMModelBuilder implements ModelBuilder {
    private SessionBuffer session ;
    private BingoTreeNode rootNode = null;
    private SVMModelListener listener = null;
    private double dThreshold = SessionBuffer.dThreshold;
    private boolean justAtLevel = SessionBuffer.getObeyLevel(); // betrachte URLs pro Baum-Level oder global
    
    //private int type;
//    private int itDepth;
    private int nodeCounter = 0;
//    private boolean bBreak = false;
//    private boolean ignoreNoSparse;
    
    public static String plusOne = "+1 ";
    public static String minusOne = "-1 ";
    
    /**
     *  Fehlermeldung: Keine IDF-Tabelle zur Feature-Selektion bzw. -Gewichtung geladen.
     */
    public static String NO_IDF = "Keine IDF/Feature-Tabelle fr: ";
    
    
    /**
     *  Konstruiert einen neuen SVMModelBuilder mit dem Vector der Trainingsdaten und einer gegebenen Hierarchie.
     *
     * @param  vURLs     Die Trainingsdaten (als Vektor von BaseURL's).
     * @param  rootNode  Der Wurzelknoten der Hierarchie.
     */
    public SVMModelBuilder() {
        session = SessionBuffer.getInstance();
        this.rootNode = session.getRootNode();
        //this.ignoreNoSparse = session.ignoreNoSparse;
        this.justAtLevel = session.getObeyLevel();// betrachte URLs pro Baum-Level oder global
        this.dThreshold = session.dThreshold;
    }
    
    public void classifyTrainingSamples() {
        if ( listener != null )
            listener.message( "Classify training samples" );
        
        ArrayList allDocs = rootNode.getPathDocuments();
        Iterator it = allDocs.iterator();
        while (it.hasNext()) {
            BingoDocument bd = (BingoDocument) it.next();
            BingoTreeNode node = bd.getNode();
            if (node != null)
                bd.setSvmRank(new Double(node.validate(bd)));
            session.log(bd.getStatus() + " : " + bd.getSvmRank(), 5);
        }
        it = null;
        rootNode.calcAvgSVMScore();
        
        if (session.dbConnect) {
            if ( listener != null )
                listener.message( "Store updated scores" );
            
            DBInterface dbinterface = DBInterface.getInstance();
            it = allDocs.iterator();
            while (it.hasNext()) {
                BingoDocument bd = (BingoDocument) it.next();
                Double svmrank = bd.getSvmRank();
                long id = bd.getId();
                String klasse = bd.getNode().label;
                dbinterface.updateSVMResult(id, klasse, svmrank);
            }
        }
        
        allDocs=null;
    }
    
    public void buildAllModels() {
        initFeatures();
        nodeCounter = 0;
        buildNodeModels(rootNode);
        
        classifyTrainingSamples();
        
        deleteFeatures();
        
        if ( listener != null )
            listener.finished( "FERTIG" );
    }
    
    public void initFeatures() {
        ArrayList allDocs = rootNode.getPathDocuments();
        Iterator it = allDocs.iterator();
        while (it.hasNext()) {
            BingoDocument bd = (BingoDocument)it.next();
            Collection featureVector = bd.getFeatureVector();
            
            bd.setMappedFeatures(rootNode.documentMapping(featureVector));
        }
        it=null;
        allDocs=null;
    }
    
    public void deleteFeatures() {
        ArrayList allDocs = rootNode.getPathDocuments();
        Iterator it = allDocs.iterator();
        while (it.hasNext()) {
            BingoDocument bd = (BingoDocument)it.next();
            bd.setMappedFeatures(null);
        }
        it=null;
        allDocs=null;
    }
    
    public void buildNodeModels(BingoTreeNode node) {
        if (node == null || node.isLeaf())
            return;
        
        Enumeration enum = node.children();
        while (enum.hasMoreElements()) {
            nodeCounter++;
            if ( listener != null )
                listener.barChanged( nodeCounter, node );
            
            BingoTreeNode child = (BingoTreeNode) enum.nextElement();
            
            if ( listener != null )
                listener.nodeChanged( child );
            
            ArrayList posDocs=null, negDocs=null;
            if (this.justAtLevel) {
                posDocs = child.getNodeDocuments();
                negDocs = child.getOppositeDocuments();
            }
            else {
                posDocs = child.getPathDocuments();
                negDocs = child.getOppositePathDocuments();
            }
            
            // Animation
            //===================================================
            
            if (posDocs == null || posDocs.size() <= 0) {
                if ( listener != null )
                    listener.warning( "no POS samples, skipped" );
                child.classifier = null;
            }
            
            if (negDocs == null || negDocs.size() <= 0) {
                if ( listener != null )
                    listener.warning( "no NEG samples, skipped" );
                child.classifier = null;
            }
            
            if (negDocs != null && posDocs != null && negDocs.size() > 0 && posDocs.size() > 0) {
                if ( listener != null )
                    listener.message( "POS: " + posDocs.size() + "   NEG: " + negDocs.size());
                
                //===================================================
                
                SVMClassifierModel model = null;
                try {
                    model = buildModel(node,posDocs,negDocs);
                    if (model != null)
                        model.setThreshold( dThreshold );
                }
                catch (Exception e) {
                    session.log(e);
                    if ( listener != null )
                        listener.handleException( e );
                }
                child.classifier = model;
                model=null;
                
                buildNodeModels(child);
            }
        }
        enum=null;
    }
    
    public SVMClassifierModel buildModel(BingoTreeNode node, ArrayList posDocs, ArrayList negDocs) throws Exception
	   {
	      if (node == null || posDocs == null || negDocs == null)
	         return null;
	      
	      if (posDocs.size() + negDocs.size() <= 0)
	         return null;
	      
	      String s=System.getProperty("file.separator");
	      String d=System.getProperty("user.dir");
	      String expdir = "temp";
	      String path = d + s + expdir + s;
	      String filename = path + "svm_" + posDocs.size() + "_" + negDocs.size();
	      Iterator it = null;
	      File file = null;
	      
	      //=======================================================
	      // Vorbereitung der Trainingsdaten
	      
	      String dat_filename = filename + ".dat";
	      file = new File(dat_filename);
	      if(file.exists())
	         file.delete();
	      FileWriter fos=new FileWriter(dat_filename);
	      BufferedWriter bos = new BufferedWriter(fos);
	      
	      ArrayList allDocs= new ArrayList(posDocs.size() + negDocs.size());
	      allDocs.addAll(posDocs);
	      allDocs.addAll(negDocs);
	      it = allDocs.iterator();
	      int posCount = 0;
	      int negCount = 0;
	      while (it.hasNext())
	      {
	         BingoDocument tmp = (BingoDocument)it.next();
	         Object[] features = node.classMapping(tmp.getMappedFeatures());
	         if (features == null || features.length <= 0)
	            continue;
	         
	         String sample = node.getLightString(features);
	         if (posDocs.contains(tmp))
	         {
	            posCount ++;
	            sample = plusOne + sample;
	         }
	         else
	         {
	            negCount ++;
	            sample = minusOne + sample;
	         }
	         bos.write(sample + "\n");
	         sample = null;
	      }
	      it = null;
	      allDocs = null;
	      
	      bos.flush(); fos.flush();
	      bos.close(); fos.close();
	      bos=null; fos=null;
	      file=null;
	      
	      //=======================================================
	      // Modellierung
	      
	      String mod_filename = filename + ".mod";
	      String out_filename = filename + ".out";
	      
	      Runtime rt = Runtime.getRuntime();
	      String command =  path + "train.bat " + path + " " + dat_filename + " " + mod_filename + " " + out_filename;
	      session.log(command, 3);
	      Process prozess = rt.exec(command);
	      
	      InputStream stderr = prozess.getInputStream();
	      InputStreamReader isr = new InputStreamReader(stderr);
	      BufferedReader br = new BufferedReader(isr);
	      String line = null;
	      //      System.out.println("<ERROR>");
	      while ( (line = br.readLine()) != null)
	      {
	         //                System.out.println(line);
	      }
	      //            System.out.println("</ERROR>");
	      line = null;
	      isr = null;
	      stderr = null;
	      
	      
	      prozess.waitFor();
	      //int res = prozess.exitValue();
	      prozess.getInputStream().close(); prozess.getOutputStream().close(); prozess.getErrorStream().close();
	      prozess = null;
	      command = null;
	      rt = null;
	      
	      out_filename = null;
	      dat_filename = null;
	      
	      //=======================================================
	      SVMClassifierModel result = SVM_Light.readModelFile(mod_filename);
	      mod_filename = null;
	      //===========================================================
	      
	      // Clean all tmp files
	      
	      rt = Runtime.getRuntime();
	      command =  path + "clean.bat " + path ;
	      prozess = rt.exec(command);
	      
	      
	      stderr = prozess.getInputStream();
	      isr = new InputStreamReader(stderr);
	      br = new BufferedReader(isr);
	      line = null;
	      //      System.out.println("<ERROR>");
	      while ( (line = br.readLine()) != null)
	      {
	         //                System.out.println(line);
	      }
	      //            System.out.println("</ERROR>");
	      line = null;
	      isr = null;
	      stderr = null;
	      
	      
	      
	      prozess.waitFor();
	      //res = prozess.exitValue();
	      
	      prozess.getInputStream().close(); prozess.getOutputStream().close(); prozess.getErrorStream().close();
	      prozess = null;
	      command = null;
	      rt = null;
	      
	      //===========================================================
	      
	      return result;
	   }
    
    public void registerListener(SVMModelListener l) {
        this.listener = l;
    }
}