/*
 *  WordNet-Java
 *
 *  Copyright 1998 by Oliver Steele.  You can use this software freely so long as you preserve
 *  the copyright notice and this restriction, and label your changes.
 */
package edu.brandeis.cs.steele.wn;
import java.util.NoSuchElementException;

/**
 *  Instances of this class enumerate the possible major syntactic categories,
 *  or <b>p</b> art's <b>o</b> f <b>s</b> peech. Each <code>POS</code> has a
 *  human-readable label that can be used to print it, and a key by which it can
 *  be looked up.
 *
 * @author     Oliver Steele, steele@cs.brandeis.edu
 * @created    14. August 2002
 * @version    1.0
 */
public class POS {
  //
  // Class variables
  //
  public final static POS NOUN = new POS( "noun", "n" );
  public final static POS VERB = new POS( "verb", "v" );
  public final static POS ADJ = new POS( "adjective", "a" );
  public final static POS ADV = new POS( "adverb", "r" );

  /**
   *  A list of all <code>POS</code>s.
   */
  public final static POS[] CATS = {NOUN, VERB, ADJ, ADV};

  //
  // Instance implementation
  //
  protected String label;
  protected String key;


  protected POS( String label, String key ) {
    this.label = label;
    this.key = key;
  }


  //
  // Object methods
  //
  public String toString() {
    return "[POS " + label + "]";
  }


  public boolean equals( Object object ) {
    return ( object instanceof POS ) && key.equals( ( (POS) object ).key );
  }


  public int hashCode() {
    return key.hashCode();
  }


  //
  // Accessor
  //
  /**
   *  Return a label intended for textual presentation.
   *
   * @return    The label value
   */
  public String getLabel() {
    return label;
  }


  /**
   *  Return the <code>PointerType</code> whose key matches <var>key</var> .
   *
   * @param  key  Description of the Parameter
   * @return      Description of the Return Value
   */
  public static POS lookup( String key ) {
    for ( int i = 0; i < CATS.length; ++i ) {
      if ( key.equals( CATS[i].key ) ) {
        return CATS[i];
      }
    }
    throw new NoSuchElementException( "unknown POS " + key );
  }
}
