package multivalent.std.adaptor.pdf;

import java.awt.Color;
import java.awt.Transparency;
import java.awt.image.*;
import java.awt.color.*;
import java.awt.geom.AffineTransform;
import java.awt.geom.Point2D;
import java.io.InputStream;
import java.io.ByteArrayInputStream;
import java.io.IOException;
import java.util.Map;
import java.util.HashMap;
import java.util.Iterator;
import javax.imageio.*;
import javax.imageio.stream.ImageInputStream;

import static multivalent.std.adaptor.pdf.COS.*;

import phelps.imageio.plugins.Fax;
import phelps.awt.color.ColorSpaceCMYK;
import phelps.awt.color.ColorSpaceLab;
import phelps.util.Arrayss;
import phelps.io.InputStreams;



/**
	Interpret PDF image types, from objects or inline (content stream):
	DCT (JPEG), CCITT FAX (Group 3, Group 3 2D, Group 4), raw samples (bit depth 1,2,4,8), JBIG2.
	Does no cacheing; always creates new image.

	@see javax.imageio.*
	@see phelps.imageio.plugins.Fax

	@version $Revision: 1.84 $ $Date: 2004/04/09 05:05:02 $
*/
public class Images {
  static final boolean DEBUG = false && multivalent.Multivalent.DEVEL;

  private static final int JPEG_CS_BILEVEL = 0;	// encodings taken from SPIFF
  private static final int JPEG_CS_YCbCr1 = 1;
  private static final int JPEG_CS_NONE = 2;
  private static final int JPEG_CS_YCbCr2 = 3;
  private static final int JPEG_CS_YCbCr3 = 4;
  private static final int JPEG_CS_GREYSCALE = 8;
  private static final int JPEG_CS_PHOTO_Ycc = 9;
  private static final int JPEG_CS_RGB = 10;
  private static final int JPEG_CS_CMY = 11;
  private static final int JPEG_CS_CMYK = 12;
  private static final int JPEG_CS_YCCK = 13;
  private static final int JPEG_CS_CIELab = 14;

  private static final byte[] JPEG_SPIFF1 = { (byte)0xff, (byte)0xd8, (byte)0xff, (byte)0xe8 },	JPEG_SPIFF2 = "SPIFF\0".getBytes(),
	JPEG_JFIF1 = { (byte)0xff, (byte)0xd8, (byte)0xff, (byte)0xe0 }, JPEG_JFIF2 = "JFIF\0".getBytes(),
	/*JPEG_ADOBE1 = { 0xee } ,*/ JPEG_ADOBE2 = "Adobe".getBytes();

  private static boolean hasJPEG2000=true, hasJBIG2=true;


  private Images() {}

  /**
	Constructs new BufferedImage from dictionary attributes and data in stream.
	@param imgdict   image XObject, or Dict with {@link STREAM_DATA} key set for inline images
  */
  public static BufferedImage createImage(Dict imgdict, InputStream in, Color fillcolor, PDFReader pdfr) throws IOException {
	assert imgdict!=null && in!=null && ("Image".equals(imgdict.get("Subtype")) || null==imgdict.get("Subtype")/*inline*/);

	BufferedImage img;  // AffineImageOp requires BufferedImage, not just java.awt.Image
	int w = pdfr.getObjInt(imgdict.get("Width")), h = pdfr.getObjInt(imgdict.get("Height"));

//long start = System.currentTimeMillis();
	String filter = getFilter(imgdict, pdfr);
//System.out.println(imgdict.get("Name")+", filter="+filter+" "+w+"x"+h);
	if ("DCTDecode".equals(filter)) img = createJPEG(imgdict, in, pdfr);
	else if ("JPXDecode".equals(filter)) img = createJPEG2000(imgdict, in);
	else if ("CCITTFaxDecode".equals(filter)) img = createFAX(imgdict, in, fillcolor, pdfr);
	else if ("JBIG2Decode".equals(filter) /*|| "no abbreviation".equals(filter)*/) img = createJBIG2(imgdict, in);
	else { /*assert filter==null: filter;--still FlateDecode*/ img = createRaw(imgdict, w, h, in, fillcolor, pdfr); }  // raw samples, including most inline images
//Object cs = pdfr.getObject(imgdict.get("ColorSpace"));
//System.out.println(filter+","+imgdict.get("BitsPerComponent")+", "+(CLASS_ARRAY==cs.getClass()? " / "+((Object[])cs)[0]: cs));
	if (img==null) return null;     // IOException, JBIG2, or problem with samples
//long end = System.currentTimeMillis();
//System.out.println("time = "+(end-start));
	//assert w==img.getWidth(): "width="+img.getWidth()+" vs param "+w;  // possible that parameters are wrong
	//X assert h==img.getHeight(): "height="+img.getHeight()+" vs param "+h; => if short of data, shrink height

	return img;
  }


  /**
	Scale and rotate according to affine transform, adjusting image origin from PDF lower-left to Java upper-left.
	@see #createImage(Dict, InputStream, Color, PDFReader)
  */
  public static BufferedImage createScaledImage(Dict imgdict, InputStream in, AffineTransform ctm, Color fillcolor, PDFReader pdfr) throws IOException {
	assert imgdict!=null && in!=null && ctm!=null && "Image".equals(imgdict.get("Subtype")) || imgdict.get("Subtype") == null/*inline*/;

	BufferedImage img = null;
	try {
		img = createImage(imgdict, in, fillcolor, pdfr);
	} catch (Exception e) {	// unchecked exceptions from MediaLib
		e.printStackTrace();
	}
	if (img==null) return null;     // JBIG2, IOException
//if (true) { System.out.println("img = "+img); return img; }

	int w = pdfr.getObjInt(imgdict.get("Width")), h = pdfr.getObjInt(imgdict.get("Height"));


	// scaling affine transform
	// "transformation from image space to user space could be described by the matrix [1/w 0 0 -1/h 0 1]."
	AffineTransform iat;
	if (ctm.getScaleX() != 0.0 && false) {
		Point2D srcpt = new Point2D.Double(w,h), transpt = new Point2D.Double();
		srcpt.setLocation(1.0, 1.0); ctm.deltaTransform(srcpt, transpt);
		System.out.println(srcpt+" => "+transpt+" in "+w+"X"+h+" "+ctm);

		double xscale=ctm.getScaleX(), yscale=ctm.getScaleY();
		iat = new AffineTransform(xscale/w, ctm.getShearY(),
			ctm.getShearX(), -/*invert*/(yscale/h + (yscale<0.0? -1.0: 1.0)/h),    // "+1.0/h" is fuzz so don't get choppy image from bad PDF distillation
			transpt.getX()>=0.0? 0.0: -transpt.getX(), transpt.getY()>=0.0? 0.0: -transpt.getY());

	} else if (ctm.getScaleX() != 0.0) {   // Math.abs(ctm.getScaleX()) > Math.abs(ctm.getShearX())
		double xscale=ctm.getScaleX(), yscale=ctm.getScaleY();
		//if (Math.abs(xscale*w)<1.0) xscale=(xscale<0.0? -1.0: 1.0); if (Math.abs(yscale*h)<1.0) yscale=(yscale<0.0? -1.0: 1.0);     // tiny images at least 1x1 => always add one pixel, so 0+1=1 min
//System.out.println("scaled "+(xscale*w)+"x"+(yscale*h));
		//double xscale=ctm.getScaleX()/w, yscale = ctm.getScaleY()/h;//(h>2? ctm.getScaleY()/h: 1.0);//ctm.getScaleY()<0? -1.0: 1.0); -- what, flip 1 pixel?
//if (ctm.getShearX()!=0.0) System.out.println("scale + shear "+ctm);
		iat = new AffineTransform(xscale/w, ctm.getShearY(), ctm.getShearX(), -/*invert*/(yscale/h + (yscale<0.0? -1.0: 1.0)/h), 0.0, 0.0);   // "+1.0/h" is fuzz so don't get choppy image from bad PDF distillation
		iat = new AffineTransform(xscale/w, 0.0, 0.0, -/*invert*/(yscale + (yscale<0.0? -1.0: 1.0))/h, 0.0, 0.0);   // Java bug with scale+shear on images of more than small size, so ignore shear
		//double xround = 1.0/w /* 0.5*/, yround = -1.0/h /* 0.5*/;     // copy images
		//iat = new AffineTransform(xscale/w + xround, ctm.getShearX(), ctm.getShearY(), /*-invert*/yscale/h + yround, 0.0, 0.0);
		if (iat.getScaleX()<0.0) iat.translate(-w,0);   // it happens
		if (iat.getScaleY()<0.0) iat.translate(0,-h);   // in transformed space
//System.out.println(ctm+": yscale="+yscale+" / height="+h+" => "+iat.getScaleY());

	} else {    // 90 or -90 degree rotation
//System.out.println(ctm);
		double xshear=ctm.getShearX(), yshear=ctm.getShearY();
		iat = new AffineTransform(0.0, yshear/w, -xshear/h, 0.0, 0.0,0.0);
		if (iat.getShearX()<0.0) iat.translate(0,-h);
		if (iat.getShearY()<0.0) iat.translate(-w,0);
//System.out.println("rotated image: "+w+"x"+h+"  "+ctm+" => "+iat);   //+", pt="+pt);
	}
//if (DEBUG) System.out.println(w+"x"+h+", "+ctm+" => "+iat);



	// final image
	//Image imgout = img;
//System.out.println(iat);
	//String filter = getFilter(imgdict, pdfr);
	// almost always have to transform, if only to invert
	//boolean isFAX = "CCITTFaxDecode".equals(getFilter(imgdict, pdfr));
/*
	if (isFAX && iat.getScaleY()<0.0) iat = new AffineTransform(iat.getScaleX(), iat.getShearY(), iat.getShearX(), -iat.getScaleY(), 0.0, 0.0);

	int aftype = iat.getType();
	if (AffineTransform.TYPE_IDENTITY==aftype || AffineTransform.TYPE_TRANSLATION==aftype) {    // -- never happens
		// nothing

	} else if (iat.getScaleY()>0.0 && (AffineTransform.TYPE_GENERAL_SCALE==aftype || AffineTransform.TYPE_UNIFORM_SCALE==aftype)) {
		// Image.getScaledInstance not faster with SMOOTH, FAST stinks, AA slow and bad, DEFAULT fast but bad.  And doesn't handle PDF's upside down images (can flip FAX).
		// faster to handle as drawImage()? (hardware?) => magnify lens not that much faster
		int neww=(int)Math.round(w * iat.getScaleX()), newh=(int)Math.round(h * iat.getScaleY());
		imgout = img.getScaledInstance(neww, newh, Image.SCALE_DEFAULT);
System.out.println("scale only: "+iat.getScaleX()+" x "+iat.getScaleY()+", img="+img);

	} else*/

	try {
		boolean fshrink = false && Math.abs(iat.getScaleX())<0.5 && Math.abs(iat.getScaleY())<0.5;
//if (fshrink) System.out.println(getFilter(imgdict, pdfr)+", "+iat+", small? "+fshrink);
		if (fshrink) {
			AffineTransformOp aop = new AffineTransformOp(iat, AffineTransformOp.TYPE_NEAREST_NEIGHBOR);	// faster but drops text within image on scale up -- buggy: Tekton tomato goes black
			img = aop.filter(img, null);

		} else if ("CCITTFaxDecode".equals(getFilter(imgdict, pdfr))) {    // custom scaling for FAX
			img = Fax.scale(img, iat);
//System.out.println("CCITT through Fax.Scale "+w+"x"+h+" * "+iat+" => "+img);

		} else if (Boolean.TRUE==pdfr.getObject(imgdict.get("ImageMask")) && Color.BLACK.equals(fillcolor) && w*h<5*1024) {	// same test as in createRaw()
			img = Fax.scale(img, iat);

		} else {
			AffineTransformOp aop = new AffineTransformOp(iat, AffineTransformOp.TYPE_BILINEAR);    // TYPE_NEAREST_NEIGHBOR faster but drops text within image on scale up
			img = aop.filter(img, null);
		}

	} catch (/*java.awt.image.{RasterFormat,ImagingOp}*/Exception ioe) {
		ioe.printStackTrace();
		//System.out.println(img); System.out.println(ctm+" => "+iat); System.out.println(img.getColorModel());
		System.out.println(ioe);
		System.err.println(imgdict.get("Name")+" "+getFilter(imgdict, pdfr)+" "+w+"X"+h+", w/"+ctm+" => "+iat);//+" => "+pt);
		if (PDF.DEBUG) System.exit(1);
		//img=null; -- return untransformed?
	}

	//assert img!=null: imgdict;  // could be too small to show
	return img;
  }


  /** Process inline image into Node. */
  public static BufferedImage createScaledInline(/*imgdict created here,*/ InputStreamComposite in, Dict csres, AffineTransform ctm, Color fillcolor, PDFReader pdfr) throws IOException {
	Dict iidict = pdfr.readInlineImage(in);

	InputStream iin = pdfr.getInputStream(iidict);
	Object csobj=iidict.get("ColorSpace"); if (csres!=null && csres.get(csobj)!=null) iidict.put("ColorSpace", csres.get(csobj));   // key not literal
	BufferedImage img = Images.createScaledImage(iidict, iin, ctm, fillcolor, pdfr);
	iin.close();

	assert img!=null: "bad INLINE IMG "+iidict; //+", len="+bout.size();    // no JBIG2 in inline
	return img;
  }



  /**
	Return image part of filter, which may be in a cascade, or <code>null</code> if none.
	Expands abbreviations ("DCT" => "DCTDecode", "CCF" => "CCITTFaxDecode").
	For example, from <code>[ASCII85Decode CCF]</code>, returns <code>CCITTFaxDecode</code>.
  */
  public static String getFilter(Dict imgdict, COSSource coss) throws IOException {
	Object attr = imgdict.get("Filter"); if (coss!=null)/*inline*/ attr = coss.getObject(attr);
	String f;
	if (attr==null || OBJECT_NULL==attr) f=null;  // raw samples, uncompressed -- found in inline images
	else if (CLASS_NAME==attr.getClass()) f = (String)attr;
	else { assert CLASS_ARRAY==attr.getClass();      // usually image filter wrapped in ASCII.  no JPG wrapping FAX I hope!
		Object[] oa=(Object[])attr;
		Object o = oa.length>0? oa[oa.length-1]: null;	// image filter must be last
		f = (String)(coss!=null? coss.getObject(o): o);
	}

	if ("DCT".equals(f)) f="DCTDecode"; else if ("CCF".equals(f)) f="CCITTFaxDecode";  // canonicalize
	if (!"DCTDecode".equals(f) && !"CCITTFaxDecode".equals(f) && !"JBIG2Decode".equals(f) && !"JPXDecode".equals(f)) f = null;	// make sure it's image

	return f;
  }

  /**
	Returns image's /DecodeParms, or <code>null</code> if none (or {@link OBJECT_NULL}).
	If /DecodeParms is an array, the one corresponding to the image is always the last array element.
  */
  public static Dict getDecodeParms(Dict imgdict, PDFReader pdfr) throws IOException {
	Object o = pdfr.getObject(imgdict.get("DecodeParms"));
	Object dp = o==null || OBJECT_NULL==o? OBJECT_NULL: 
		CLASS_DICTIONARY==o.getClass()? o: 
		/*assert CLASS_ARRAY==o.getClass()?*/ pdfr.getObject(((Object[])o)[((Object[])o).length-1/*better be last*/]);	// 07001 has /DecodeParms [null null]

	return dp!=OBJECT_NULL? (Dict)dp: null;
  }

  /** Returns file type suffix corresponding to PDF <var>filter</var>, e.g., PDF <code>DCTDecode</code> returns <code>jpg</code>. */
  public static String getSuffix(String filter) {
	return "DCTDecode".equals(filter)? "jpg":
		"JPXDecode".equals(filter)? "jp2":
		"CCITTFaxDecode".equals(filter)? "fax":
		"JBIG2Decode".equals(filter)? "jbig2":
		"raw";
  }


  /** Hand off to ImageIO. */
  static BufferedImage createJPEG(Dict imgdict, InputStream in, PDFReader pdfr) throws IOException {
	assert imgdict!=null && in!=null;

	// If 4-color, have to distinguish between YCCK and CMYK.
	// Don't want to rely on sun.* class, so parse JPEG, which means we need two copies of data: one for IIO to consume and another for Adobe APP14.
	ColorSpace cs = pdfr.getColorSpace(imgdict.get("ColorSpace"), null, null);
	int jcs = JPEG_CS_YCbCr3;
	int nComp = cs.getNumComponents();
	if (nComp == 4 /*|| true*/) {
		Object o = imgdict.get("Length");
		byte[] data = InputStreams.toByteArray(in, o!=null? pdfr.getObjInt(o): 10*1024);
		jcs = getJPEGTransform(data);
		in = new ByteArrayInputStream(data);
	}

	BufferedImage img;
	//img = ImageIO.read(in); -- easiest, but already know it's JPEG
	ImageReader iir = (ImageReader)ImageIO.getImageReadersByFormatName("JPEG").next();
	ImageIO.setUseCache(false);
	ImageInputStream iin = ImageIO.createImageInputStream(in);
	iir.setInput(iin, true);   //new MemoryCacheImageInputStream(in));
	try {
//	Object o = pdfr.getObject(imgdict.get("ColorSpace")), o2 = CLASS_ARRAY==o.getClass()? ((Object[])o)[0]: o;
//System.out.println("colorspace: "+o+"/"+o2+" => "+cs);
		final ColorSpace RGB = ColorSpace.getInstance(ColorSpace.CS_sRGB);
//if (DEBUG) System.out.println("JPEG, colorspace: "+cs+" vs RGB "+ColorSpace.getInstance(ColorSpace.CS_sRGB)+", jcs="+jcs);
		if (nComp==4 || cs instanceof ColorSpaceLab/*jcs==JPEG_CS_CMYK/*cs != RGB /*||true/*&& jcs!=JPEG_CS_YCbCr3*/) {	// AffineImageOp can't handle non-RGB colorspace, so convert data into RGB
			ImageReadParam irp = iir.getDefaultReadParam(); 
			//irp.setSourceRenderSize(new Dimension(1000,1000)); => can't for JPEG
			//irp.setSourceSubsampling(2,2, 0,0); => not what we want
			Raster r = convertJPEG(iir.readRaster(0, irp/*null*/), jcs, cs);	// don't store ref so can gc?
			ColorModel cm = new ComponentColorModel(RGB, false, true, Transparency.OPAQUE, DataBuffer.TYPE_BYTE);
			img = new BufferedImage(cm, (WritableRaster)r, true, null);
			//ColorConvertOp cco = new ColorConvertOp(ColorSpace.getInstance(ColorSpace.CS_LINEAR_RGB), ColorSpace.getInstance(ColorSpace.CS_LINEAR_RGB), null);
			//img = cco.filter(img, null); => dog slow!

		} else {	// YCbCr, grayscale, RGB
			img = iir.read(0, null);
		}

	} catch (IOException e) { img = null; System.err.println("Couldn't read JPEG: "+e); }

	iir.dispose();
	iin.close();

	return img;
  }

  /**
	Java's ImageIO can't process 4-component images and Java2D can't apply AffineTransformOp either, so convert raster data to RGB.
	Technique due to MArk Stephens.
  */
  private static Raster convertJPEG(Raster r, int jcs, ColorSpace cs) {
	int width = r.getWidth(), height = r.getHeight();
	byte[] rgb = new byte[width*height*3];
	boolean fAdobe = jcs<0; jcs = Math.abs(jcs);
//System.out.println("nComp = "+cs.getNumComponents());	// 1, 3, 4
if (DEBUG) System.out.println("converting JPEG "+jcs+" / Adobe? "+fAdobe);

	// image can be big and samples at 4 bytes each across CMYK = 70M for 2100x2100 image, so convert in horizontal chunks
	// If conversion done internally by IIO, could access samples directly and avoid allocation and copying.
	final int HUNK = 50;
	Map<Integer,Integer> cache = new HashMap<Integer,Integer>(20000);	// 30 sec => 9 sec on DisneyMap.pdf
long start = System.currentTimeMillis();
	for (int x=0,y=0,w=width, base=0; y<height; y+=HUNK) {
		int h = Math.min(height-y, HUNK);

		if (cs.getNumComponents()==3) {	// L*a*b* seen in colour.pdf, DisneyMap.pdf
			int[] A = r.getSamples(x,y,w,h, 0, (int[])null), B = r.getSamples(x,y,w,h, 1, (int[])null), C = r.getSamples(x,y,w,h, 2, (int[])null);
			float[] abc = new float[3];
			for (int i=0,imax=A.length; i<imax; i++, base+=3) {
				//if (a==A[i] && b==B[i] && c==C[i] && i>0) { rgb[base]=rgb[base-3]; rgb[base+1]=rgb[base-3+1]; rgb[base+2]=rgb[base-3+2]; }	// 40=>30 sec on DisneyMap.pdf => wash with hash of all conversions
				int a=A[i], b=B[i], c=C[i];
				Integer key = new Integer((a<<16) | (b<<8) | c);
				Integer val = cache.get(key);
				if (val!=null) {
					int v = val.intValue();
					rgb[base] = (byte)(v>>16); rgb[base+1] = (byte)(v>>8); rgb[base+2] = (byte)v;
				} else {
					abc[0]=a*100f/255f; abc[1]=b-128f; abc[2]=c-128f;
					float[] def = cs.toRGB(abc);
//if (i<5) System.out.println(A[i]+" "+B[i]+" "+C[i]+" => "+def[0]+" "+def[1]+" "+def[2]);
					int di = (int)(def[0]*255f), ei = (int)(def[1]*255f), fi = (int)(def[2]*255f);
					rgb[base] = (byte)di; rgb[base+1] = (byte)ei; rgb[base+2] = (byte)fi;
					cache.put(key, new Integer((di<<16) | (ei<<8) | fi));
				}
			}
			A = B = C = null;

		} else if (JPEG_CS_YCCK==jcs) {
//if (DEBUG && y==0) System.out.println("converting YCCK JPEG to RGB");
			float[] Y = r.getSamples(x,y,w,h, 0, (float[])null), Cb = r.getSamples(x,y,w,h, 1, (float[])null), Cr = r.getSamples(x,y,w,h, 2, (float[])null), K = r.getSamples(x,y,w,h, 3, (float[])null);
			//ColorSpace YCC = ColorSpace.getInstance(ColorSpace.CS_PYCC);
			//float[] ycc = new float[3];
			for (int i=0,imax=Y.length; i<imax; i++, base+=3) {
				// faster to track last cmyk and save computations on stretches of same color?
				float yy = Y[i] - K[i], cb=Cb[i], cr=Cr[i];
				float val = yy + 1.402f*(cr-128f); rgb[base] = val<=0.0? (byte)0: val>=255.0? (byte)0xff: (byte)(val+0.5f);
				val = yy - 0.34414f*(cb-128f) - 0.71414f*(cr-128f); rgb[base+1] = val<=0.0? (byte)0: val>=255.0? (byte)0xff: (byte)(val+0.5f);
				val = yy + 1.772f * (cb-128f); rgb[base+2] = val<=0.0? (byte)0: val>=255.0? (byte)0xff: (byte)(val+0.5f);
				/*
				ycc[0] = Y[i] - K[i]; ycc[1]=Cb[i]; ycc[2]=Cr[i];
				float[] cvt = YCC.toRGB(ycc);
				rgb[i] = (byte)cvt[0]; rgb[i+1] = (byte)cvt[1]; rgb[i+2] = (byte)cvt[2];
				*/
			}
			Y = Cb = Cr = K = null;

		} else { assert JPEG_CS_CMYK==jcs: jcs;
//*if (DEBUG && y==0)*/ System.out.println("converting CMYK JPEG to RGB");
			int[] C = r.getSamples(x,y,w,h, 0, (int[])null), M = r.getSamples(x,y,w,h, 1, (int[])null), Y = r.getSamples(x,y,w,h, 2, (int[])null), K = r.getSamples(x,y,w,h, 3, (int[])null);
			for (int i=0,imax=C.length; i<imax; i++, base+=3) {
				int k = K[i];
				rgb[base] = (byte)(255 - Math.min(255, C[i] + k));
				rgb[base+1] = (byte)(255 - Math.min(255, M[i] + k));
				rgb[base+2] = (byte)(255 - Math.min(255, Y[i] + k));
			}
			C = M = Y = K = null;
		}
	}

long end = System.currentTimeMillis(); if (DEBUG) System.out.println("elapsed " +((end-start)/1000)+", cache size = "+cache.size());
	if (fAdobe) {
//System.out.println("inverting Photoshop data");
		//for (int i=0,imax=rgb.length; i<imax; i++) rgb[i] = (byte)(255 - (rgb[i]&0xff));
	}

	// from other image types we know InterleavedRaster's can be manipulated by AffineTransformOp, so create one of those.
//if (DEBUG && rgb.length > 10*1024*1024) { System.gc(); System.out.println("free = "+Runtime.getRuntime().freeMemory()); }
	r = null;	// gc
	return Raster.createInterleavedRaster(new DataBufferByte(rgb, rgb.length), width, height, width*3,3, new int[] {0,1,2}, null);
  }

  private static int getJPEGTransform(byte[] data) {
	int jcs = JPEG_CS_YCbCr2;	// JPEG_CS_YCCK;

	if (Arrayss.equals(data,0, JPEG_SPIFF1,0, JPEG_SPIFF1.length) && Arrayss.equals(data,4, JPEG_SPIFF2,0, JPEG_SPIFF2.length)) {
		jcs = data[0x18] & 0xff;
PDF.sampledata("SPIFF "+jcs);

/* can have JFIF and Adobe APP14
	} else if (Arrayss.equals(data,0, JPEG_JFIF1,0, JPEG_JFIF1.length) && Arrayss.equals(data,4, JPEG_JFIF2,0, JPEG_JFIF2.length)) {
		jcs = JPEG_CS_YCbCr2;
PDF.sampledata("JFIF "+jcs);*/

	} else {	// scan for Adobe
		for (int i=0,imax=data.length-2; i<imax; ) {
			assert (data[i]&0xff)==0xff;
			int type = data[i+1] & 0xff;	// want unsigned bytes!
			//out_.println("+"+i+": "+Integer.toHexString(type)/*+", len="+len*/);
			i += 2;	// 0xff and type

			if (type==0x01 || (0xd0 <= type&&type <= 0xda)) {

			} else if (type==0xda) {
				i = i + ((data[i]&0xff)<<8) + (data[i+1]&0xff);
				while (true) {
					for ( ; i<imax; i++) if ((data[i]&0xff)==0xff && data[i+1]!=0) break;
					int rst = data[i+1]&0xff;
					if (0xd0 <= rst&&rst <= 0xd7) i+=2; else break;
				}
			} else {
				/*if (0xc0 <= type&&type <= 0xcf) {	// SOF
					Nf = data[i+7] & 0xff;	// 1=grayscale, 3=YCbCr, 4=YCCK or CMYK
					jcs = Nf==1? JPEG_CS_GREYSCALE: Nf==3? JPEG_CS_YCbCr2: Nf==4? JPEG_CS_CMYK;
					if (JPEG_CS_GREYSCALE==jcs || JPEG_CS_YCbCr2==jcs) break;
				} else*/ if (type == 0xee) {	// APP14: possible Adobe
					if (Arrayss.equals(data,i+2, JPEG_ADOBE2,0, 5)) {
						int xform = data[i+13];	// 0=CMYK, 1=YCbCr, 2=YCCK
						jcs = xform==1? JPEG_CS_YCbCr2: xform==2? JPEG_CS_YCCK: JPEG_CS_CMYK;
//if (DEBUG) System.out.println("Adobe, xform="+xform+" => "+jcs);
						//jcs = -jcs;	// mark as Adobe APP so invert due to Photoshop bug
						break;
					}
				}
				i = i + ((data[i]&0xff)<<8) + (data[i+1]&0xff);
			}
		}
	}

	return jcs;
  }



  /** Hand off to ImageIO. */
  static BufferedImage createJPEG2000(Dict imgdict, InputStream in) throws IOException {
	assert /*imgdict!=null &&*/ in!=null;
	BufferedImage img;

	//img = ImageIO.read(in); -- easiest, but already know it's JPEG
	if (!hasJPEG2000) return null;
	Iterator<ImageReader> i = ImageIO.getImageReadersByFormatName("JPEG2000");
	if (!i.hasNext()) { hasJPEG2000 = false; System.err.println("No decoder for JPEG2000 -- install ImageIO from Sun or Apple"); return null; }
	ImageReader iir = i.next();
	ImageIO.setUseCache(false);
	ImageInputStream iin = ImageIO.createImageInputStream(in);
	iir.setInput(iin, true);   //new MemoryCacheImageInputStream(in));
	try {
		img = iir.read(0);
	} catch (IOException e) { img = null; }
	iir.dispose();
	iin.close();

	return img;
  }



  /** Decode parameters from PDF dictionary and pass on to {@link phelps.imageio.plugins.Fax}. */
  static BufferedImage createFAX(Dict imgdict, InputStream in, Color fillcolor, PDFReader pdfr) throws IOException {
	assert imgdict!=null && in!=null/* && pdfr!=null -- testing*/;

//System.out.println(imgdict.get("Name")+", K="+K+", cols="+cols+", rows="+rows+", swap="+swapbw);
	Dict dp = getDecodeParms(imgdict, pdfr); if (dp==null) new Dict(3);   // no /DP very rare, at least have /K -1

	Object o;
	int K = (o=dp.get("K"))!=null? pdfr.getObjInt(o): 0;   // <0 = pure 2D Group 4 (only one used in practice by PDF), ==0 = 1D Group 3, >0 = mixed 1D/2D Group 3
if (K>0) { PDF.sampledata("Group 3 mixed");	if (DEBUG) System.exit(0); }
	int cols = (o=dp.get("Columns"))!=null? pdfr.getObjInt(o): 1728;  assert cols>=1 /*&& cols%8==0 -- not root_cz.pdf by Ghostscript 7.05*/: cols;     // rounds up to nearest byte boundary
	int width = pdfr.getObjInt(imgdict.get("Width"));		// can be less than /Columns
	if (cols != width) { int mod = width%8; if (mod!=0) width += (8 - mod); }	// round up to multiple of 8, as needed by pdfdb/000208.pdf
	int height = pdfr.getObjInt(imgdict.get("Height"));		// Estimate rows by height, not cols*1.5 as can get "FAX strips" of 2500 width but only 107 height, so hugely overestimate and provoke flood of garbage collections!
	int rows = (o=dp.get("Rows"))!=null? pdfr.getObjInt(o): -height/*just go until hit EOF -- estimate for performance*/;  //assert rows>=1;   //if (rows<=0) rows=Integer.MAX_VALUE;

	boolean EndOfLine = Boolean.TRUE == pdfr.getObject(dp.get("EndOfLine"));	// defaults to false and only Boolean.TRUE makes it true
	boolean EndOfBlock = Boolean.FALSE != pdfr.getObject(dp.get("EndOfBlock"));	// defaults to true and only Boolean.FALSE makes it false
	boolean EncodedByteAlign = Boolean.TRUE == pdfr.getObject(dp.get("EncodedByteAlign"));
//if (EncodedByteAlign) PDF.sampledata("FAX /EncodedByteAlign");
	boolean BlackIs1 = Boolean.TRUE == pdfr.getObject(dp.get("BlackIs1"));
	int DamagedRowsBeforeError = (o=dp.get("DamagedRowsBeforeError"))!=null? pdfr.getObjInt(o): 0;

	//o=pdfr.getObject(imgdict.get("Decode")); => handled in scale()'s color map
	boolean swapbw = java.util.Arrays.equals(Normalize.A10, (Object[])pdfr.getObject(imgdict.get("Decode")));

	final byte white = (byte)(BlackIs1^swapbw? 1: 0);     // /Decode [1 0] different than BlackIs1.  INVERTED for performance: set to 0 as Java clears page to 0 and most of page is white, so save rewriting; then swap in indexed color map.


	//Runtime rt = Runtime.getRuntime(); System.out.println("total = "+rt.totalMemory()+" .. "+rt.maxMemory());
	//if (2 * cols * Math.abs(rows) + 5*1024*1024 > rt.maxMemory()) { System.err.println("Fax too big: "+cols+"x"+Math.abs(rows)+" > "+rt.maxMemory()); return null; }	// seen 8500x8500
	BufferedImage img = Fax.decode(K, width, cols, rows,  EndOfLine, EndOfBlock, EncodedByteAlign, white, DamagedRowsBeforeError,  in);


	// if image masks replace color map
	if (Boolean.TRUE == pdfr.getObject(imgdict.get("ImageMask"))) {
//System.out.println("image mask "+fillcolor);// imgdict);
		ColorModel cm = fillcolor.getRed()<8 && fillcolor.getBlue()<8 && fillcolor.getBlue()<8?	// seen r=0,g=0,b=1
			Fax.GRAY8_REV:
			new IndexColorModel(8, 2, new int[] { 0, fillcolor.getRGB() }, 0, true, 0/*trans*/, DataBuffer.TYPE_BYTE);
		img = new BufferedImage(cm, img.getRaster(), false, new java.util.Hashtable());
	}

	return img;
  }



  /** Not implemented. */
  static BufferedImage createJBIG2(Dict imgdict, InputStream in) throws IOException {
	if (!hasJBIG2) return null;
	hasJBIG2=false; System.err.println("No decoder for JBIG2."); return null;
  }



  /**
	Create image from raw samples, in various bit depths (8, 4, 2, 1), in a variety of color spaces, with various numbers of samples per component (4,3,1).
  */
  static BufferedImage createRaw(Dict imgdict, int w, int h, InputStream in, Color fillcolor, PDFReader pdfr) throws IOException {
	assert imgdict!=null && w>0 && h>0 && in!=null && fillcolor!=null && pdfr!=null;

	// 1. colorspace
	Boolean fmask = (Boolean)pdfr.getObject(pdfr.getObject(imgdict.get("ImageMask")));
	int bpc = fmask==Boolean.TRUE? 1: pdfr.getObjInt(imgdict.get("BitsPerComponent"));
	ColorModel cm = createRawColorModel(pdfr.getObject(imgdict.get("ColorSpace")), fmask, bpc, fillcolor, pdfr);
//System.out.println("colorspace = "+imgdict.get("ColorSpace")+" => "+cm+" "+w+"x"+h);
	int spd = (cm instanceof IndexColorModel? 1 : cm.getNumComponents());   // 1 for gray or indexed (regardless of cm.getNumComponents()), 3 for rgb, 4 for CMYK


	// 2. data, tweaked for /Decode and CMYK=>RGB
//int est = (w*h * spd*bpc + (bpc-1))/8;
//System.out.println("data buf len = "+bout.size());
	byte[] rawdata = InputStreams.toByteArray(in, w*h*spd);
//System.out.println(est+" vs "+rawdata.length+" actual @ "+bpc);

	// check that have enough data
	// if don't, reduce height.  Ghostscript 5.50 reports one line too high, as on netzwerke.pdf
	int scanline = (w*spd * bpc + 7) / 8, reqlen = scanline * h;
//System.out.println("RAW: "+w+"x"+h+" @ "+bpc+" = "+(w*h)+", req="+reqlen);
	if (rawdata.length < reqlen) {	// show data that do have
if (DEBUG) { System.out.println("short data: "+rawdata.length+" < "+reqlen+": "+w+"x"+h+" * "+spd+" @ "+bpc+" bpp"); System.out.println(imgdict); }
		//byte[] req = new byte[reqlen]; System.arraycopy(rawdata,0, req,0, rawdata.length); rawdata = req;
		h = rawdata.length / scanline;
	}

	// decode matrix
	Object[] decode = (Object[])pdfr.getObject(imgdict.get("Decode"));
	if (decode!=null) {  // cheaper to tweak color map if indexed
//System.out.println("decode = "+o);
		// LATER: if IndexColorModel, twiddle map instead
		double[] da = new double[decode.length];
		for (int i=0,imax=decode.length; i<imax; i++) da[i] = ((Number)decode[i]).doubleValue();
		if (da.length==2 && da[0]==1.0 && da[1]==0.0) { // || (da[2]==1.0 && da[3]==0.0 && da[))) {
//System.out.println("invert");
			for (int i=0,imax=rawdata.length; i<imax; i++) rawdata[i] ^= 0xff;
		}// else PDF.unsupported("Decode matrix: "+da.length+" len, "+da);
	}

/*
	if (bpc==8 && cm.getColorSpace() instanceof ICC_ColorSpace) {
		int[] hist = new int[256];
		for (int i=0,imax=rawdata.length; i<imax; i++) hist[rawdata[i]&0xff]++;
		for (int i=0,imax=hist.length; i<imax; i++) if (hist[i]>0) System.out.print(i+"="+hist[i]+" ");
	}*/

	// Work around Java AffineTransformOp limitation on 4-component color spaces by transcoding data to RGB.  Just handles 8-bit CMYK case now, which is all that I've seen.
	if (bpc==8 && cm.getColorSpace() instanceof ColorSpaceCMYK) {
		byte[] newdata = new byte[w * h * 3];
		for (int i=0,imax=Math.min(w*h*4, rawdata.length),j=0; i<imax; i+=4, j+=3) {	// sometimes extra data
			int k = rawdata[i+3]&0xff;
			newdata[j]   = (byte)(255 - Math.min(255, (rawdata[i]&0xff)   + k));
			newdata[j+1] = (byte)(255 - Math.min(255, (rawdata[i+1]&0xff) + k));
			newdata[j+2] = (byte)(255 - Math.min(255, (rawdata[i+2]&0xff) + k));
		}
		rawdata = newdata;
		cm = new ComponentColorModel(ColorSpace.getInstance(ColorSpace.CS_sRGB), false, true, Transparency.OPAQUE, DataBuffer.TYPE_BYTE);
		spd = 3;
	}

	// special case for Type 3: convert data to Fax for good looking scaling: 8x larger, but worth it
	WritableRaster r;
	if (Boolean.TRUE==fmask && Color.BLACK.equals(fillcolor) && w*h<5*1024 /*&& fType3*/) {
		assert spd==1: spd;
//System.out.println("converting black ImageMask for Fax scaling: "+w+"x"+h+", bpc="+bpc);
		//byte[] newdata = new byte[(w+1)*(h+1)];	// Fax scaling requires extra row and column
		byte[] newdata = new byte[w*h];
		for (int y=0, i=0, base=0; y<h; y++, base+=w/*+1*/) {
			int valid=0, bits=0;
			for (int x=0; x<w; x++) {
				if (valid<=0) { bits=rawdata[i++]&0xff; valid=8; }
				newdata[base+x] = (bits&0x80)==0? (byte)1: 0;
				bits <<= 1; valid--;
			}
		}

		rawdata = newdata;
		bpc = 8;
		fmask = Boolean.FALSE;
		cm = new ComponentColorModel(ColorSpace.getInstance(ColorSpace.CS_GRAY), false, true, Transparency.OPAQUE, DataBuffer.TYPE_BYTE);
		r = Raster.createInterleavedRaster(new DataBufferByte(rawdata, rawdata.length), w,h, w/*+1/*stride*/,1, new int[] { 0 }, null);

	} else r = createRawRaster(rawdata, w, h, bpc, spd);
	//WritableRaster r = createRawRaster(rawdata, w, h, bpc, spd);



	// 3. colorspace + raster = image
	// create image with color model so colors survive affine transform
	BufferedImage img = null;
try {
	img = new BufferedImage(cm, r, false, new java.util.Hashtable());
} catch (Exception e) {
	//e.printStackTrace();
	System.err.println(e);

	System.out.println("color model = "+cm);
	SampleModel sm = r.getSampleModel();
	System.out.println("sample model = "+sm);
/*
		if (sm instanceof ComponentSampleModel) {
			if (sm.getNumBands() != getNumComponents()) {
				return false;
			}
			for (int i=0; i<nBits.length; i++) {
				if (sm.getSampleSize(i) < nBits[i]) {
					return false;
				}
			}
			return (raster.getTransferType() == transferType);
		}
*/
	// from IndexedColorModel.java
	//int size = r.getSampleModel().getSampleSize(0);
	//System.out.println(r.getTransferType()+" ==? "+cm.getTransferType());
	//System.out.println(r.getNumBands()+" ==? 1");
	//System.out.println((1 << size)+" >=? "+((IndexColorModel)cm).getMapSize());
	//if (DEBUG) System.exit(1);

	System.out.println("sample model instance of ComponentSampleModel "+(sm instanceof ComponentSampleModel));
	System.out.println("num bands = "+sm.getNumBands()+" ==? "+cm.getNumComponents()+" cm num comp");
	int[] nbits = cm.getComponentSize();
	for (int i=0; i<nbits.length; i++) System.out.println("  "+sm.getSampleSize(i)+" >=? "+nbits[i]);
	System.out.println(r.getTransferType()+" ==? "+sm.getTransferType());

}

	return img;
  }


  private static ColorModel createRawColorModel(Object csref, Boolean fmask, int bpc, Color fillcolor, PDFReader pdfr) throws IOException {
	//assert csref!=null; // => image mask
	Object csobj = pdfr.getObject(csref);
	//ColorSpace cs = (csobj!=null/*error, except for image masks*/? getColorSpace(csref): ColorSpace.getInstance(ColorSpace.CS_sRGB));
	ColorSpace cs = pdfr.getColorSpace(csref, null, null);  assert cs!=null || Boolean.TRUE==fmask;

	ColorModel cm;

	// 1: INDEX or GRAY
	if (Boolean.TRUE==fmask) {   // as abused in riggs.pdf
		assert bpc == 1;
//System.out.println("fillcolor = "+fillcolor+"/"+Integer.toHexString(fillcolor.getRGB())+", decode="+decode+", "+w+"x"+h);
		// "unmasked areas will be painted using the current nonstroking color"
		cm = new IndexColorModel(1, 2, new int[] { fillcolor.getRGB(), 0 }, 0, /*false--grr*/true, 1, DataBuffer.TYPE_BYTE);
		//cm = new IndexColorModel(1, 2, new int[] { Color.BLUE.getRGB(), Color.GREEN.getRGB() }, 0, false, -1, DataBuffer.TYPE_BYTE);

	} else if (CLASS_ARRAY==csobj.getClass() && ("Indexed".equals(((Object[])csobj)[0]) || "I".equals(((Object[])csobj)[0]))) { // indexed special case
		Object[] oa = (Object[])csobj;   // [/Indexed base hival samples]
		int hival = pdfr.getObjInt(oa[2]);
		Object cmap = pdfr.getObject(oa[3]);  assert CLASS_DICTIONARY==cmap.getClass() || CLASS_STRING==cmap.getClass();

		byte[] samp;
		if (CLASS_DICTIONARY==cmap.getClass()) {
			samp = pdfr.getStreamData(oa[3]/*NOT cmap-- could be encrypted*/, false, false);
//System.out.println("index data "+oa[3]+"/"+pdfr.getObject(oa[3])+" => size="+samp.length+" vs hival="+hival+"/"+((hival+1)*3));

		} else { assert CLASS_STRING==cmap.getClass();
			StringBuffer sb = (StringBuffer)cmap;
			samp = new byte[sb.length()];
			for (int i=0,imax=sb.length(); i<imax; i++) samp[i] = (byte)sb.charAt(i);
		}

		// Java's IndexColorModel requires base RGB space, so convert colors from current color space
		Object base = pdfr.getObject(oa[1]);     // "array or name"
		if (null==base || "DeviceRGB".equals(base) || "RGB".equals(base)) {}  // ok as is
		else {  // "any device or CIE-based color space or (in PDF 1.3) a Separation or DeviceN space, but not a Pattern space or another Indexed space."
			ColorSpace bcs = pdfr.getColorSpace(base, null, null);
			int spd = bcs.getNumComponents(), sampcnt = samp.length / spd;
			byte[] brgb = new byte[sampcnt*3];  // RGB
			float[] fsamp = new float[spd];
//System.out.println("converting "+base+"/"+oa[1]+"/"+bcs+" to RGB, length "+samp.length+"=>"+brgb.length+", spd="+spd+", sampcnt="+sampcnt);
			for (int i=0, is=0, id=0; i<sampcnt; i++) {
				for (int j=0; j<spd; j++) fsamp[j] = (samp[is++]&0xff) / 255f;  // GIVE US UNSIGNED BYTES!
				float[] rgb = bcs.toRGB(fsamp);
//if (id<10*spd || i>250) System.out.println(fsamp[0]+" "+fsamp[1]+" "+fsamp[2]+" "+fsamp[3]+" => "+rgb[0]+" "+rgb[1]+" "+rgb[2]);
				for (int j=0; j<3; j++) brgb[id++] = (byte)(rgb[j] * 255f);
//if (i<10 || i>250) System.out.println(i+": "+fsamp[0]+" "+fsamp[1]+" "+fsamp[2]+" "+fsamp[3]+" => "+brgb[id-3]+" "+brgb[id-2]+" "+brgb[id-1]);
			}
			samp = brgb;
		}
//for (int i=0,imax=samp.length; i<imax; i+=3) System.out.println((i/3)+": "+Integer.toHexString(samp[i]&0xff)+" "+Integer.toHexString(samp[i+1]&0xff)+" "+Integer.toHexString(samp[i+2]&0xff));
//System.out.println("bpc = "+bpc+", hival="+hival);

		// cache conversion...

		// Java bug: IndexColorModel.isCompatibleRaster() checks (1<<raster_bpc)>=map_size should be (1<<raster_bpc)<=map_size
		//assert b.length == hival+1: bpc+" vs "+(hival+1);
		// workaround Java bug: make equal (ColorModels not cached).  Tickled by pdfTeX-0.14d.  RESTORE.
//System.out.println("indexed size = "+(hival+1)+", bpc="+bpc+", samp.length="+samp.length);
		cm = new IndexColorModel(bpc, Math.min(1<<bpc, hival+1), samp, 0, false);
		// Index works with both Interleaved and Packed Rasters
		// shanghai one color for transparency in case of /Mask

	} else if (cs.getNumComponents()==1 && bpc<8) {   // grayscale (/Indexed handled above)
		// Java has to have IndexColorModel for Packed Rasters so translate grayscale
		if (bpc==1) cm = new IndexColorModel(1, 2, new int[] { 0x000000, 0xffffff }, 0, false, -1, DataBuffer.TYPE_BYTE);
		else if (bpc==2) cm = new IndexColorModel(2, 4, new int[] { 0x000000, 0x404040, 0xc0c0c0, 0xffffff }, 0, false, -1, DataBuffer.TYPE_BYTE);
		else /*assert bpc==4*/ cm = new IndexColorModel(4, 16, new int[] { 0x000000, 0x111111, 0x222222, 0x333333, 0x444444, 0x555555, 0x666666, 0x777777, 0x888888, 0x999999, 0xaaaaaa, 0xbbbbbb, 0xcccccc, 0xdddddd, 0xeeeeee, 0xffffff }, 0, false, -1, DataBuffer.TYPE_BYTE);


	// 3/4: RGB, CMYK
	} else if (bpc==8 || bpc==4) {    // 4-bit case split out in readRawData to separate bytes
		cm = new ComponentColorModel(cs, false, true, Transparency.OPAQUE, DataBuffer.TYPE_BYTE);   // "each sample in a separate data element", that is, interleaved raster

	} else {    // fabricate an IndexColorModel so can pack all samples into single data element, 2^(2*4) = 256, but usually 2^(2*3)=64
		//ColorSpace cs = (csobj!=null/*error, except for image masks*/? getColorSpace(csref): ColorSpace.getInstance(ColorSpace.CS_sRGB));
		int spd = cs.getNumComponents();
		assert (spd==3 || spd==4) && (bpc==2 || bpc==1): "bpc="+bpc;
//System.out.println("imgdict = "+imgdict+", colorspace="+csobj+", spd="+spd);

		// color, 3 or 4 components -- compute entries vis-a-vis base colorspace
		int po2 = (1<<bpc);
		byte[] b = new byte[(1<<(bpc*spd)) * spd];
//System.out.println("bpc="+bpc+", spd="+spd+", b.length="+b.length);
		float red=0f, green=0f, blue=0f, black=0f, inc=(float)(1f/po2);
		float[] comp = new float[4];    // enough for CMYK
		Color c;
		int i=0;
		for (int r=0; r<po2; r++, red+=inc, green=0f, blue=0f, black=0f) {
			comp[0]=red;
			for (int g=0; g<po2; g++, green+=inc, blue=0f, black=0f) {
				comp[1]=green;
				for (int bl=0; bl<po2; bl++, blue+=inc, black=0f) {
					comp[2]=blue;
					for (int k=0; k<po2; k++, black+=inc) {
						comp[3]=black;
						c = new Color(cs, comp, 0f);
						b[i++]=(byte)c.getRed(); b[i++]=(byte)c.getGreen(); b[i++]=(byte)c.getBlue();
						if (spd==3) break;
					}
				}
			}
		}
		assert i==b.length: i+" vs "+b.length;

		cm = new IndexColorModel(bpc, b.length/spd, b, 0, false);
	}

	return cm;
  }


  private static WritableRaster createRawRaster(byte[] rawdata, int w, int h, int bpc, int spd) {
	WritableRaster r;

	int[] offs=new int[spd]; for (int i=0; i<spd; i++) offs[i]=i;

//System.out.println("bpc="+bpc+", spd="+spd+", w="+w+", h="+h);
//if (bpc<8) System.out.println("w="+w+", h="+h+", bpc="+bpc+", spd="+spd+", data buf len = "+rawdata.length+" vs w*h*spd="+(w*h*spd*bpc/8+w*spd));
	// provided SampleModels either force all samples in same element (packed), or sample-per-element (interleaved)
	if (bpc==8 || (bpc==2 && spd==4)) {
		// if (spd==1) PackedRaster; else
//System.out.println("data len ="+rawdata.length+" vs  needed "+(w*h*spd)+"  ("+(rawdata.length/(w*spd))+" rows)");
		r = Raster.createInterleavedRaster(new DataBufferByte(rawdata, rawdata.length), w, h, w*spd,spd, offs, null);

	} else if (spd==1) { assert bpc==4 || bpc==2 || bpc==1;   // indexed or grayscale: samples fit in a byte
//System.out.println("spd=1, bpc="+bpc+", rawdata.length="+rawdata.length+", mask="+Integer.toBinaryString((1<<bpc)-1)+", shift="+(8-bpc));//+", bands="+r.getNumBands());
		r = Raster.createPackedRaster(new DataBufferByte(rawdata, rawdata.length), w,h, bpc, null);
//System.out.print("spd==1: "); for (int i=0, imax=Math.min(10, rawdata.length); i<imax; i++) System.out.print(Integer.toHexString(rawdata[i])+" ");
//System.out.println("   @ "+ctm.getTranslateX()+","+ctm.getTranslateY()+", scale="+ctm.getScaleX()+"x"+ctm.getScaleY());
/*		if (!(cm instanceof IndexColorModel)) { // Java has to have IndexColorModel for this Raster class and got here from grayscale
			cm = (bpc==1? GRAY1: bpc==2? GRAY2: /*bpc==4* / GRAY4);
		}*/

	} else if (bpc==4 || (bpc==1 && spd==4)) {  // split two pairs of 4-bit data per byte into single 4-bits of data per byte.  Doubles size for RGB or CMYK samples, which is acceptable expansion
		int scanline = w * spd;
		boolean fodd = (scanline&1)==1;
		byte[] newdata = new byte[scanline * h + (fodd? 1: 0)];
//System.out.println("4 bpc x 2 bytes: "+w+"x"+h+", len="+rawdata.length+" => "+newdata.length+", spd="+spd+", offs="+offs);
		for (int y=0, i=0, j=0; y<h; y++) {
			for (int x=0; x<scanline; x+=2) {
				byte b = rawdata[i++];
				// decode array...
				newdata[j++] = (byte)(b&0xf0);
				newdata[j++] = (byte)((b<<4)&0xf0);
			}
			if (fodd) j--;
		}
		r = Raster.createInterleavedRaster(new DataBufferByte(newdata, newdata.length), w,h, scanline,spd, offs, null);

	} else if (bpc==2) {	// spd==3: samples smeared across data elements
		byte[] newdata = new byte[w*h];
		int valid=0, vbpc=0;
		for (int y=0, newi=0, base=0, stride=(w*spd*bpc + 7)/8; y<h; y++, base=y*stride) {
			for (int x=0; x<w; x++) {   // slow, but unusual
				if (valid<6) { vbpc = (vbpc<<8) | rawdata[base++]; valid+=8; }
				newdata[newi++] = (byte)((vbpc >> (valid-6)) & 0x3f);
				valid -= 6;
			}
		}
		//r = Raster.createPackedRaster(new DataBufferByte(newdata, newdata.length), w,h, w, new int[] { 0x30, 0x0c, 0x03 }, null);
		r = Raster.createInterleavedRaster(new DataBufferByte(newdata, newdata.length), w,h, w*spd,spd, new int[] { 0 }, null);
PDF.sampledata("2 bpc packed BYTE: "+rawdata.length+" => "+newdata.length);

	} else { assert bpc==1;
		byte[] newdata = new byte[w*h];
		int valid=0, vbpc=0;
		for (int y=0, newi=0, base=0, stride=(w*spd*bpc + 7)/8; y<h; y++, base=y*stride) {
			for (int x=0; x<w; x++) {   // slow, but unusual
				if (valid<3) { vbpc = (vbpc<<8) | rawdata[base++]; valid+=8; }
				newdata[newi++] = (byte)((vbpc >> (valid-3)) & 7);
				valid -= 3;
			}
		}
		//r = Raster.createPackedRaster(new DataBufferByte(newdata, newdata.length), w,h, w, new int[] { 4,2,1 }, null);
		r = Raster.createInterleavedRaster(new DataBufferByte(newdata, newdata.length), w,h, w*spd,spd, new int[] { 0 } , null);
PDF.sampledata("1 bit packed byte: "+rawdata.length+" => "+newdata.length);
	}

	return r;
  }
}
