package jodd.db;

import java.sql.Array;
import java.sql.Connection;
import java.sql.PreparedStatement;
import java.sql.ResultSet;
import java.sql.SQLException;
import java.sql.Statement;

import jodd.db.pool.ConnectionPool;

/**
 * Smart-and-slick SQL facade.
 * Class is <b>not</b> fool-proof and threadsafe.
 *
 * @2do Add more <code>PreparedStatement</code> setXXX() methods
 */
public class SqlUtil {

	// ---------------------------------------------------------------- properties

	private String[] debugSql;
	private String[] debugSqlParams;
	private Connection conn = null;
	private Statement st = null;				// because of speed optimization...
	private PreparedStatement pst = null;		// instanceof and castings are avoid
	private ResultSet resultSet = null;

	// ---------------------------------------------------------------- constructors

	private ConnectionPool pool = null;
	
	private Connection providedConnection = null;

	/**
	 * Default constructor from ConnectionPool.
	 *
	 * @param cp     connection pool instance
	 */
	public SqlUtil(ConnectionPool cp) {
		pool = cp;
	}
	/**
	 * Default constructor from common Connection.
	 *
	 * @param conenction opened db connection
	 */
	public SqlUtil(Connection conenction) {
		providedConnection = conenction;
	}

	// ---------------------------------------------------------------- init

	/**
	 * Sets and prepares a SQL command by using forward only and read only
	 * prepared statement.
	 *
	 * @param sql    sql query string.
	 *
	 * @exception SQLException
	 * @see #setSql(String, int, int)
	 */
	public void setSql(String sql) throws SQLException {
		setSql(sql, ResultSet.TYPE_FORWARD_ONLY, ResultSet.CONCUR_READ_ONLY);
	}

	/**
	 * Sets and prepares a SQL command by using prepared statements.
	 * <p>
	 *
	 * It first closes existing resultset, if any availiable. Then it
	 * closes existing prepared statement. After, obtains connection
	 * from database if not already taken. Sets commit mode, if necessary.
	 *
	 * @param sql        sql query string.
	 * @param resultType result type
	 * @param resultSetConcurrency
	 *                   result concurrency
	 *
	 * @exception SQLException
	 */
	public void setSql(String sql, int resultType, int resultSetConcurrency) throws SQLException {
		if (resultSet != null) {
			resultSet.close();
			resultSet = null;
		}
		if (pst != null) {
			pst.close();
			pst = null;
		}
		if (conn == null) {
			if (pool != null) {
				conn = pool.getConnection();
			} else {
				conn = providedConnection;
			}
		}
		setCommitMode();
		pst = conn.prepareStatement(sql, resultType, resultSetConcurrency);
		createDebugSql(sql);
	}

	/**
	 * Sets and prepares a SQL command by using forward only and read only
	 * statement.
	 *
	 * @param sql    sql query string.
	 *
	 * @exception SQLException
	 * @see #setStaticSql(String, int, int)
	 */
	public void setStaticSql(String sql) throws SQLException {
		setSql(sql, ResultSet.TYPE_FORWARD_ONLY, ResultSet.CONCUR_READ_ONLY);
	}


	/**
	 * Sets and prepares a SQL command by using simple statements.
	 * <p>
	 *
	 * It first closes existing resultset, if any availiable. Then it
	 * closes existing statement. After, obtains connection
	 * from database if not already taken. Sets commit mode, if necessary.
	 *
	 * @param sql        sql query string.
	 * @param resultType result type
	 * @param resultSetConcurrency
	 *                   concurency type
	 *
	 * @exception SQLException
	 */
	public void setStaticSql(String sql, int resultType, int resultSetConcurrency) throws SQLException {
		if (resultSet != null) {
			resultSet.close();
			resultSet = null;
		}
		if (st != null) {
			st.close();
			st = null;
		}
		if (conn == null) {
			if (pool != null) {
				conn = pool.getConnection();
			} else {
				conn = providedConnection;
			}
		}
		setCommitMode();
		st = conn.createStatement(resultType, resultSetConcurrency);
		debugSql = new String[] { sql };
		debugSqlParams = new String[] {""};
	}

	// ---------------------------------------------------------------- execute

	/**
	 * Executes SELECT statements, but works with all statements. Previously closes
	 * existing result set.
	 *
	 * @return ResultSet
	 * @exception SQLException
	 */
	public ResultSet executeQuery() throws SQLException {
		if (resultSet != null) {
			resultSet.close();
			resultSet = null;
		}
		if (st != null) {
			resultSet = st.executeQuery(debugSql[0]);
			return resultSet;
		}
		if (pst != null) {
			resultSet = pst.executeQuery();
			return resultSet;
		}
		return null;
	}

	/**
	 * Executes UPDATE, INSERT or DELETE statements.
	 *
	 * @return result of executeUpdate()
	 * @exception SQLException
	 */
	public int executeUpdate() throws SQLException {
		if (st != null) {
			return st.executeUpdate(debugSql[0]);
		}
		if (pst != null) {
			return pst.executeUpdate();
		}
		return -1;
	}


	// ---------------------------------------------------------------- close

	/**
	 * Closes result set and the statement. Connection is not closed.
	 *
	 * @param rs     ResultSet created by executeQuery(), or null if ResultSet is not being used.
	 *
	 * @see #close()
	 * @deprecated use close() instead
	 */
	public void close(ResultSet rs) {
		try {
			if (rs != null) {
				rs.close();
			}
		} catch (SQLException sex) {
		}

		try {
			if (pst != null) {
				pst.close();
			}
		} catch (SQLException sex) {
		} finally {
			pst = null;
		}

		try {
			if (st != null) {
				st.close();
			}
		} catch (SQLException sex) {
		} finally {
			st = null;
		}
	}

	/**
	 * Closes result set (if previously created) and the statement. Connection is not closed.
	 */
	public void close() {
		// result set
		try {
			if (resultSet != null) {
				resultSet.close();
			}
		} catch (SQLException sex) {
		} finally {
			resultSet = null;
		}

		// prepared statement
		try {
			if (pst != null) {
				pst.close();
			}
		} catch (SQLException sex) {
		} finally {
			pst = null;
		}

		// statement
		try {
			if (st != null) {
				st.close();
			}
		} catch (SQLException sex) {
		} finally {
			st = null;
		}
	}

	/**
	 * Closes result set, preparead statements, sets default commit mode for
	 * connection and returns connection to db pool.
	 *
	 * @param rs     ResultSet created by executeQuery(), or null if ResultSet is not being used.
	 *
	 * @deprecated use closeAll() instead.
	 */
	public void closeAll(ResultSet rs) {
		close(rs);
		resetCommitMode();
		if (conn != null) {
			if (pool != null) {
				pool.freeConnection(conn);
			} else {
				try {
					conn.close();
				} catch (SQLException e) {
				}
			}
			conn = null;			
		}
	}

	/**
	 * Closes result set (if created), statements and returns connection to db
	 * pool when no ResultSets are used.
	 */
	public void closeAll() {
		close();
		resetCommitMode();
		if (conn != null) {
			if (pool != null) {
				pool.freeConnection(conn);
			} else {
				try {
					conn.close();
				} catch (SQLException e) {
				}
			}
			conn = null;			
		}
	}

	// ---------------------------------------------------------------- transactions

	private static Boolean defaultAutoCommit = null;
	/**
	 * Sets default auto-commit that SqlUtil <b>must</b> provide before executing
	 * new statement, neverness if <code>setAutoCommit</code> is used or not. By
	 * using this method and setting default auto-commit value to be the same as
	 * it is for database, it is guaranteed that if it is not explicitly set
	 * different, the default mode will be used.
	 *
	 * <p>
	 * If default auto-commit mode is not used, connection will be examined on
	 * first usage, and that commit status will be taken as the default one.
	 *
	 * @param autoCommit default auto-commit mode
	 */
	public static void setDefaultAutoCommit(boolean autoCommit) {
		defaultAutoCommit = Boolean.valueOf(autoCommit);
	}
	/**
	 * Returns default auto-commit mode or <code>null</code> if nothing set.
	 *
	 * @return default auto-commit mode
	 */
	public static Boolean getDefaultAutoCommit() {
		return defaultAutoCommit;
	}


	private Boolean storedAutoCommit = null;			// stored auto commit for the current session

	private Boolean currentAutoCommit = null;			// current value of auto-commit

	/**
	 * Sets this connection's auto-commit mode to the given state. If a
	 * connection is in auto-commit mode, then all its SQL statements will be
	 * executed and committed as individual transactions. Otherwise, its SQL
	 * statements are grouped into transactions that are terminated by a call to
	 * either the method <code>commit</code> or the method <code>rollback</code>.
	 *
	 * <p>
	 * If this method is not used, then default database commit mode is used,
	 * and this class doesn't change anything.
	 *
	 * <p>
	 * NOTE: if this method is called <b>before</b> very first
	 * <code>setSql()</code> or <code>setStaticSql()</code>, than it will just
	 * 'remember' the commit state that will be actually set in those commands,
	 * after connection is retrieved from database. This is prefered method.
	 *
	 * <p>
	 * If it is called after, then this method will directly change connection
	 * commit mode, with all consequences.
	 *
	 * @param autoCommit true to enable auto-commit mode; false to disable it
	 *
	 * @exception SQLException
	 */
	public void setAutoCommit(boolean autoCommit) throws SQLException {
		currentAutoCommit = Boolean.valueOf(autoCommit);
		setCommitMode();
	}

	/**
	 * Retrieves the current auto-commit mode for this object.
	 * If <code>null</code> is returned then default database mode is used.
	 *
	 * @return the current auto-commit mode
	 */
	public Boolean getAutoCommit() {
		return currentAutoCommit;
	}

	/**
	 * Returns the stored auto-commit mode for the connection. This value should
	 * match database initial commit mode. If <code>null</code> is returned, then
	 * connection is not yet obtained. Therefore, this method has to be called
	 * after <code>setSql</code> methods.
	 *
	 * @return default commit mode
	 */
	public Boolean getStoredAutoCommit() {
		return storedAutoCommit;
	}

	/**
	 * Intelligently sets the commit mode. When called for the first time on
	 * availiable connection it stores default values.
	 *
	 * @exception SQLException
	 */
	private void setCommitMode() throws SQLException {
		if (conn == null) {
			return;
		}
		// remember default
		if (storedAutoCommit == null) {									// very first time on this connection
			if (defaultAutoCommit == null) {
				storedAutoCommit = Boolean.valueOf(conn.getAutoCommit());	// remember default auto-commit mode...
			} else {
				storedAutoCommit = defaultAutoCommit;						// ...or use specified default mode
			}
		}
		// set new if required
		boolean newMode;
		if (currentAutoCommit != null) {						// is there a new setting?
			newMode = currentAutoCommit.booleanValue();
		} else {
			newMode = storedAutoCommit.booleanValue();
		}
		if (newMode != conn.getAutoCommit()) {					// If new mode is different then current mode..
			conn.setAutoCommit(newMode);						// ...then specify new, specified setting
		}
	}

	/**
	 * Resets connection commit mode to default, when connection is returned back
	 * to database pool.
	 */
	private void resetCommitMode() {
		if (conn == null) {
			return;
		}
		try {
			if (currentAutoCommit != null) {
				boolean currentMode = conn.getAutoCommit();				// if current mode...
				boolean storedMode = storedAutoCommit.booleanValue();	// ...and stored mode...
				if (currentMode != storedMode) {						// ...are not equals...
					conn.setAutoCommit(storedMode);						// ...set mode to remembered default mode
				}
			}
		} catch (SQLException sex) {
		} finally {
			currentAutoCommit = null;
			storedAutoCommit = null;
		}
	}


	/**
	 * Makes all changes made since the previous commit/rollback permanent and
	 * releases any database locks currently held by this object. This method
	 * should be used only when auto-commit mode has been disabled.
	 *
	 * @exception SQLException
	 */
	public void commit() throws SQLException {
		conn.commit();
	}

	/**
	 * Undoes all changes made in the current transaction and releases any
	 * database locks currently held by this object. This method should be used
	 * only when auto-commit mode has been disabled.
	 */
	public void rollback() {
		try {
			conn.rollback();
		} catch (SQLException sex) {
		}
	}

	// ---------------------------------------------------------------- debug sql

	private static boolean debugMode = false;
	/**
	 * Sets debug mode.
	 *
	 * @param b
	 */
	public static void setDebugMode(boolean b) {
		debugMode = b;
	}
	/**
	 * Returns current debug mode.
	 *
	 * @return current debug mode
	 */
	public static boolean getDebugMode() {
		return debugMode;
	}
	
	private void createDebugSql(String sql) {
		if (getDebugMode() == false) {
			debugSql = new String[] {sql};
			return;
		}
		debugSql = sql.split("\\?");
		debugSqlParams = new String[debugSql.length];
		int i = 0;
		int limit = debugSql.length - 1;
		while (i < limit) {
			debugSqlParams[i] = "?";
			i++;
		}
		if (sql.endsWith("?")) {
			debugSqlParams[i] = "?";
		} else {
			debugSqlParams[i] = "";
		}
	}

	private void setDebugSql(int ndx, String s) {
		if (getDebugMode() == false) {
			return;
		}
		debugSqlParams[ndx - 1] = s;
	}

	private void setDebugSql(int ndx, int i) {
		if (getDebugMode() == false) {
			return;
		}
		debugSqlParams[ndx - 1] = Integer.toString(i);
	}

	private void setDebugSql(int ndx, boolean i) {
		if (getDebugMode() == false) {
			return;
		}
		debugSqlParams[ndx - 1] = Boolean.toString(i);
	}

	private void setDebugSql(int ndx, long i) {
		if (getDebugMode() == false) {
			return;
		}
		debugSqlParams[ndx - 1] = Long.toString(i);
	}

	private void setDebugSql(int ndx, byte i) {
		if (getDebugMode() == false) {
			return;
		}
		debugSqlParams[ndx - 1] = Byte.toString(i);
	}

	private void setDebugSql(int ndx, double i) {
		if (getDebugMode() == false) {
			return;
		}
		debugSqlParams[ndx - 1] = Double.toString(i);
	}

	private void setDebugSql(int ndx, float i) {
		if (getDebugMode() == false) {
			return;
		}
		debugSqlParams[ndx - 1] = Float.toString(i);
	}



	/**
	 * Returns formated pseudo-SQL string. When <code>PreparedStatement</code>
	 * is used, all ? are replaced with String representation of assigned values.
	 * Just for debugging purposes.
	 *
	 * @return formated pseudo-SQL string
	 */
	public String toString() {
		if (getDebugMode() == false) {
			return debugSql[0];
		}
		StringBuffer result = new StringBuffer();
		int i = 0;
		while (i < debugSql.length) {
			result.append(debugSql[i]);
			result.append(debugSqlParams[i]);
			i++;
		}
		return result.toString();
	}

	// ---------------------------------------------------------------- prepared statement (primitives)


	/**
	 * Sets value of int parameter in sql string.
	 *
	 * @param ndx    1-based index od parameter
	 * @param value  int value
	 *
	 * @exception SQLException
	 */
	public void setInt(int ndx, int value) throws SQLException {
		pst.setInt(ndx, value);
		setDebugSql(ndx, value);
	}

	/**
	 * Sets value of boolean parameter in sql string.
	 *
	 * @param ndx    1-based index of parameter
	 * @param value  parameter value
	 *
	 * @exception SQLException
	 */
	public void setBoolean(int ndx, boolean value) throws SQLException {
		pst.setBoolean(ndx, value);
		setDebugSql(ndx, value);
	}

	/**
	 * Sets value of long parameter in sql string.
	 *
	 * @param ndx    1-based index of parameter
	 * @param value  parameter value
	 *
	 * @exception SQLException
	 */
	public void setLong(int ndx, long value) throws SQLException {
		pst.setLong(ndx, value);
		setDebugSql(ndx, value);
	}

	/**
	 * Sets value of byte parameter in sql string.
	 *
	 * @param ndx    1-based index of parameter
	 * @param value  parameter value
	 *
	 * @exception SQLException
	 */
	public void setByte(int ndx, byte value) throws SQLException {
		pst.setByte(ndx, value);
		setDebugSql(ndx, value);
	}

	/**
	 * Sets value of double parameter in sql string.
	 *
	 * @param ndx    1-based index of parameter
	 * @param value  parameter value
	 *
	 * @exception SQLException
	 */
	public void setDouble(int ndx, double value) throws SQLException {
		pst.setDouble(ndx, value);
		setDebugSql(ndx, value);
	}


	/**
	 * Sets value of float parameter in sql string.
	 *
	 * @param ndx    1-based index of parameter
	 * @param value  parameter value
	 *
	 * @exception SQLException
	 */
	public void setFloat(int ndx, float value) throws SQLException {
		pst.setFloat(ndx, value);
		setDebugSql(ndx, value);
	}


	// ---------------------------------------------------------------- prepared statement (objects)

	/**
	 * Sets value of String parameter in sql string.
	 *
	 * @param ndx    1-based index of parameter
	 * @param value  parameter value
	 *
	 * @exception SQLException
	 */
	public void setString(int ndx, String value) throws SQLException {
		if (value != null) {
			pst.setString(ndx, value);
		} else {
			pst.setNull(ndx, java.sql.Types.VARCHAR);
		}
		setDebugSql(ndx, value);
	}


	/**
	 * Sets value of java.sql.Date parameter in sql string.
	 *
	 * @param ndx    1-based index of parameter
	 * @param value  parameter value
	 *
	 * @exception SQLException
	 */
	public void setDate(int ndx, java.sql.Date value) throws SQLException {
		if (value != null) {
			pst.setDate(ndx, value);
		} else {
			pst.setNull(ndx, java.sql.Types.DATE);
		}
		setDebugSql(ndx, value.toString());
	}

	/**
	 * Sets value of java.sql.Timestamp parameter in sql string.
	 *
	 * @param ndx    1-based index of parameter
	 * @param value  parameter value
	 *
	 * @exception SQLException
	 */
	public void setTimestamp(int ndx, java.sql.Timestamp value) throws SQLException {
		if (value != null) {
			pst.setTimestamp(ndx, value);
		} else {
			pst.setNull(ndx, java.sql.Types.TIMESTAMP);
		}
		setDebugSql(ndx, value.toString());
	}

	/**
	 * Sets value of Array parameter in sql string.
	 *
	 * @param ndx    1-based index of parameter
	 * @param value  parameter value
	 *
	 * @exception SQLException
	 */
	public void setArray(int ndx, Array value) throws SQLException {
		if (value != null) {
			pst.setArray(ndx, value);
		} else {
			pst.setNull(ndx, java.sql.Types.ARRAY);
		}
		setDebugSql(ndx, value.toString());
	}

	
	/**
	 * Sets value of BigDecimal parameter in sql string.
	 *
	 * @param ndx    1-based index of parameter
	 * @param value  parameter value
	 *
	 * @exception SQLException
	 */
	public void setBigDecimal(int ndx, java.math.BigDecimal value) throws SQLException {
		if (value != null) {
			pst.setBigDecimal(ndx, value);
		} else {
			pst.setNull(ndx, java.sql.Types.NUMERIC);
		}
		setDebugSql(ndx, value.toString());
	}

	/**
	 * Sets value of Blob parameter in sql string.
	 *
	 * @param ndx    1-based index of parameter
	 * @param value  parameter value
	 *
	 * @exception SQLException
	 */
	public void setBlob(int ndx, java.sql.Blob value) throws SQLException {
		if (value != null) {
			pst.setBlob(ndx, value);
		} else {
			pst.setNull(ndx, java.sql.Types.BLOB);
		}
		setDebugSql(ndx, value.toString());
	}

	/**
	 * Sets value of Clob parameter in sql string.
	 *
	 * @param ndx    1-based index of parameter
	 * @param value  parameter value
	 *
	 * @exception SQLException
	 */
	public void setClob(int ndx, java.sql.Clob value) throws SQLException {
		if (value != null) {
			pst.setClob(ndx, value);
		} else {
			pst.setNull(ndx, java.sql.Types.CLOB);
		}
		setDebugSql(ndx, value.toString());
	}


}
