package jodd.file.filters;

import java.io.File;
import java.io.FileFilter;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

import jodd.util.StringFlags;
import jodd.util.StringUtil;


/**
 * FileFilter that matches files with use of regular expression.
 * By default, both files and directories are matched, but only
 * by their name (not including path). This behaviour can be changed
 * by using the following flags:
 * <ul>
 *   <li>p - match also the path part of the file name</li>
 *   <li>d - do not match directories (skip)</li>
 * </ul>
 *
 * Here are some tips for regular expressions:
 * <ul>
 *   <li>.* : matches any number of character</li>
 *   <li>.? : matches zero or one character</li>
 * </ul>
 *
 * Note: when path is also used for matching (flag p), it is assumed that
 * file separators are unix-like, to prevent complications. This is done
 * because <code>File</code> understands unix-like file separators, no matter
 * what underlying system really is (at least it works so on windows).
 * Therefore, regexp patterns should written for unix-like file names.
 */
public class RegExpFileFilter implements FileFilter {

	// ---------------------------------------------------------------- flags

	private static StringFlags flags = new StringFlags();

	static {
		flags.addFlag('p', "matchPath", true);
		flags.addFlag('d', "skipDirs", true);
	}

	private boolean matchPath;

	/**
	 * Set match path flag.
	 *
	 * @param v
	 */
	public void setMatchPath(boolean v) {
		matchPath = v;
	}

	public boolean getMatchPath() {
		return matchPath;
	}

	private boolean skipDirs;
	public void setSkipDirs(boolean v) {
		skipDirs = v;
	}
	public boolean getSkipDirs() {
		return skipDirs;
	}
	
	// ---------------------------------------------------------------- init

	/**
	 * Regular Expression file filter.
	 *
	 * @param regexp regexp pattern
	 * @param opt    subset of following flags:
	 *               <ul>
	 *               <li>p - match also the path part of the file name</li>
	 *               <li>d - do not match directories, i.e. all directories matches ok.</li>
	 *               </ul>
	 */
	public RegExpFileFilter(String regexp, String opt) {
		flags.parse(opt, this);
		if (regexp != null) {
			pattern = Pattern.compile(regexp);
		} else {
			pattern = null;
		}
	}


	/**
	 * Regular Expression file filter with default behaviour.
	 *
	 * @param regexp regexp pattern
	 */
	public RegExpFileFilter(String regexp) {
		this(regexp, null);
	}



	// ---------------------------------------------------------------- FileFilter

	private Pattern pattern = null;

	private static boolean isUnix = false;
	static {
		if (File.separatorChar == '/') {
			isUnix = true;
		} else {
			isUnix = false;
		}
	}
	
	/**
	 * Tests whether or not the specified File matches conditions.
	 *
	 * @param f      file to be tested
	 *
	 * @return <code>true</code> if file matches ok, otherwise <code>false</code>
	 */
	public boolean accept(File f) {
		if (f.isDirectory()) {					// file is a directory
			if (skipDirs == true) {				// return if no matching required
				return true;
			}
		}
		if (pattern == null) {					// no pattern specified -> no matching possible
			return true;						// therefore file is ok
		}
		String file = null;
		if (matchPath == false) {
			file = f.getName();					// do not match the path
		} else {
			file = f.getPath();					// match complete file name
			if (isUnix == false) {
				file = StringUtil.replace(file, File.separatorChar, '/');
			}
		}
		Matcher matcher = pattern.matcher(file);	// regexp match
		return matcher.matches();
	}
}
