package jodd.servlet;

import java.io.File;
import java.io.IOException;
import java.util.HashMap;

import javax.servlet.ServletException;
import javax.xml.parsers.DocumentBuilder;
import javax.xml.parsers.DocumentBuilderFactory;
import javax.xml.parsers.ParserConfigurationException;
import javax.xml.transform.TransformerException;

import org.apache.xpath.XPathAPI;
import org.w3c.dom.Document;
import org.w3c.dom.NamedNodeMap;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;
import org.xml.sax.SAXException;

/**
 * ActionController helper class.
 */
public class ActionControllerUtil {

	/**
	 * Default relative address of main configuration xml file that defines
	 * actions.
	 */
	public final static String defaultConfigFile = "WEB-INF/actions.xml";

	/**
	 * Parses actions configuration xml file and loads a hashmap with processed
	 * data. Target HashMap will be used by ActionController for redirecting the
	 * requests.
	 * <p>
	 *
	 * This method is recursive, i.e. configuration file may contains imports of
	 * other configuration files. This may be used to split one big configuration
	 * file to many shorter ones. The result will be the same.
	 *
	 * @param dest     target hashmap
	 * @param global_forwards
	 *                 target global forwards ActionData
	 * @param filePath local root-path of the folder that contains all config files. usually it
	 *                 is the local path of the application web folder.
	 * @param fileName relative path to the exact actions xml configuration file.
	 *
	 * @exception ServletException
	 */
	public static void parseFile(HashMap dest, ActionData global_forwards, String filePath, String fileName) throws ServletException {

		// build file path for configuration file
		if (fileName == null) {
			fileName = defaultConfigFile;
		}
		String configFile =	filePath;
		if (filePath.endsWith("/") == false) {
			configFile += "/";
		}
		if (fileName.startsWith("/")) {
			fileName = fileName.substring(1);
		}
		configFile += fileName;

		// read configuration XML
		try {
			// create a DocumentBuilderFactory and configure it
			DocumentBuilderFactory dbf = DocumentBuilderFactory.newInstance();
			dbf.setNamespaceAware(true);
			dbf.setValidating(false);
			dbf.setIgnoringComments(true);

			// create a DocumentBuilder and parse the input file to get a Document object
			DocumentBuilder db = dbf.newDocumentBuilder();
			Document doc = db.parse(new File(configFile));

			// read all actions
			NodeList nodelist = XPathAPI.selectNodeList(doc, "/actions/action");
			for (int i = 0; i < nodelist.getLength(); i++) {
				Node action_node = nodelist.item(i);
				NamedNodeMap nmm = action_node.getAttributes();
				String path = null;
				String type = null;
				String method = null;
				for (int j = 0; j < nmm.getLength(); j++) {
					Node action_attr = nmm.item(j);
					if (action_attr.getNodeType() != Node.ATTRIBUTE_NODE) {
						continue;
					}
					if (action_attr.getNodeName().equals("path")) {
						path = action_attr.getNodeValue();
					}
					if (action_attr.getNodeName().equals("type")) {
						type = action_attr.getNodeValue();
					}
					if (action_attr.getNodeName().equals("method")) {
						method = action_attr.getNodeValue();
					}
				}
				ActionData actd = new ActionData(path, type, method);
				
				NodeList action_forwards = action_node.getChildNodes();
				for (int k = 0; k < action_forwards.getLength(); k++) {
					Node forward_node = action_forwards.item(k);
					if (forward_node.getNodeType() != Node.ELEMENT_NODE) {
						continue;
					}
					if (forward_node.getNodeName().equals("forward") == true) {
						addForward(actd, forward_node);
						continue;
					}
					if (forward_node.getNodeName().equals("param") == true) {
						addParameter(actd, forward_node);
						continue;
					}
				}
				dest.put(path, actd);
			}

			// read all global forwards
			nodelist = XPathAPI.selectNodeList(doc, "/actions/global-forwards/forward");
			for (int i = 0; i < nodelist.getLength(); i++) {
				Node forward_node = nodelist.item(i);
				addForward(global_forwards, forward_node);
			}

			// process all included files
			nodelist = XPathAPI.selectNodeList(doc, "/actions/include");
			for (int i = 0; i < nodelist.getLength(); i++) {
				Node include_node = nodelist.item(i);
				NamedNodeMap include_attrs = include_node.getAttributes();
				for (int j = 0; j < include_attrs.getLength(); j++) {
					Node attr = include_attrs.item(j);
					if (attr.getNodeType() != Node.ATTRIBUTE_NODE) {
						continue;
					}
					if (attr.getNodeName().equals("file")) {
						parseFile(dest, global_forwards, filePath, attr.getNodeValue());
						break;
					}
				}
			}
		} catch (ParserConfigurationException pcex) {
			throw new ServletException("parser configuration error" + pcex.toString());
		} catch (SAXException saxex) {
			throw new ServletException("xml parse error" + saxex.toString());
		} catch (TransformerException tex) {
			throw new ServletException("xml transforming error" + tex.toString());
		} catch (IOException ioex) {
			throw new ServletException("io error" + ioex.toString());
		}
	}


	/**
	 * Adds forward information into the ActionDate object, from a XML node.
	 *
	 * @param acd
	 * @param forward_node
	 */
	public static void addForward(ActionData acd, Node forward_node) {
		String f_name = null;
		String f_path = null;
		String f_rdrc = "false";
		NamedNodeMap forward_attributes = forward_node.getAttributes();
		for (int i = 0; i < forward_attributes.getLength(); i++) {	// read all attributes
			Node fattr = forward_attributes.item(i);
			if (fattr.getNodeType() != Node.ATTRIBUTE_NODE) {
				continue;
			}
			if (fattr.getNodeName().equals("name")) {
				f_name = fattr.getNodeValue();
			}
			if (fattr.getNodeName().equals("path")) {
				f_path = fattr.getNodeValue();
			}
			if (fattr.getNodeName().equals("redirect")) {
				f_rdrc = fattr.getNodeValue();
			}
		}
		if (f_name != null) {
			acd.putForwardPath(f_name, f_path, f_rdrc);				// add forward info to action data
		}
		return;
	}


	/**
	 * Adds parameter information into the ActionDate object, from a XML node.
	 *
	 * @param acd
	 * @param parameter_node
	 */
	public static void addParameter(ActionData acd, Node parameter_node) {
		String f_name = null;
		String f_value = null;
		NamedNodeMap parameter_attributes = parameter_node.getAttributes();
		for (int i = 0; i < parameter_attributes.getLength(); i++) {	// read all attributes
			Node fattr = parameter_attributes.item(i);
			if (fattr.getNodeType() != Node.ATTRIBUTE_NODE) {
				continue;
			}
			if (fattr.getNodeName().equals("name")) {
				f_name = fattr.getNodeValue();
			}
			if (fattr.getNodeName().equals("value")) {
				f_value = fattr.getNodeValue();
			}
		}
		if (f_name != null) {
			acd.putParameter(f_name, f_value);							// add parameter info to action data
		}
		return;
	}

}
